from __future__ import (
    absolute_import,
    division,
    print_function,
)

from pcs.test.tools.pcs_unittest import TestCase

import json
import base64
try:
    # python 2
    from urlparse import parse_qs as url_decode
except ImportError:
    # python 3
    from urllib.parse import parse_qs as url_decode

from pcs.test.tools.pcs_unittest import mock, skip
from pcs.test.tools.assertions import (
    assert_report_item_list_equal,
    assert_raise_library_error,
)
from pcs.test.tools.custom_mock import MockLibraryReportProcessor

from pcs.common import report_codes
from pcs.common.node_communicator import RequestTarget
from pcs.lib.node import NodeAddresses, NodeAddressesList
from pcs.lib.errors import LibraryError, ReportItemSeverity as Severities
from pcs.lib.external import NodeCommunicator, NodeConnectionException
import pcs.lib.booth.sync as lib


def to_b64(string):
    return base64.b64encode(string.encode("utf-8")).decode("utf-8")

@skip("TODO: rewrite for pcs.lib.communication.booth.BoothSaveFiles")
@mock.patch("pcs.lib.booth.sync.parallel_nodes_communication_helper")
class SyncConfigInCluster(TestCase):
    def setUp(self):
        self.mock_communicator = mock.MagicMock(spec_set=NodeCommunicator)
        self.mock_reporter = MockLibraryReportProcessor()
        self.node_list = NodeAddressesList(
            [NodeAddresses("node" + str(i) for i in range(5))]
        )

    def test_without_authfile(self, mock_parallel):
        lib.send_config_to_all_nodes(
            self.mock_communicator,
            self.mock_reporter,
            self.node_list,
            "cfg_name",
            "config data"
        )
        mock_parallel.assert_called_once_with(
            lib._set_config_on_node,
            [
                (
                    [
                        self.mock_communicator,
                        self.mock_reporter,
                        node,
                        "cfg_name",
                        "config data",
                        None,
                        None
                    ],
                    {}
                )
                for node in self.node_list
            ],
            self.mock_reporter,
            False
        )
        assert_report_item_list_equal(
            self.mock_reporter.report_item_list,
            [(
                Severities.INFO,
                report_codes.BOOTH_CONFIG_DISTRIBUTION_STARTED,
                {}
            )]
        )

    def test_skip_offline(self, mock_parallel):
        lib.send_config_to_all_nodes(
            self.mock_communicator,
            self.mock_reporter,
            self.node_list,
            "cfg_name",
            "config data",
            skip_offline=True
        )
        mock_parallel.assert_called_once_with(
            lib._set_config_on_node,
            [
                (
                    [
                        self.mock_communicator,
                        self.mock_reporter,
                        node,
                        "cfg_name",
                        "config data",
                        None,
                        None
                    ],
                    {}
                )
                for node in self.node_list
                ],
            self.mock_reporter,
            True
        )
        assert_report_item_list_equal(
            self.mock_reporter.report_item_list,
            [(
                Severities.INFO,
                report_codes.BOOTH_CONFIG_DISTRIBUTION_STARTED,
                {}
            )]
        )

    def test_with_authfile(self, mock_parallel):
        lib.send_config_to_all_nodes(
            self.mock_communicator,
            self.mock_reporter,
            self.node_list,
            "cfg_name",
            "config data",
            authfile="/my/auth/file.key",
            authfile_data="authfile data".encode("utf-8")
        )
        mock_parallel.assert_called_once_with(
            lib._set_config_on_node,
            [
                (
                    [
                        self.mock_communicator,
                        self.mock_reporter,
                        node,
                        "cfg_name",
                        "config data",
                        "/my/auth/file.key",
                        "authfile data".encode("utf-8")
                    ],
                    {}
                )
                for node in self.node_list
                ],
            self.mock_reporter,
            False
        )
        assert_report_item_list_equal(
            self.mock_reporter.report_item_list,
            [(
                Severities.INFO,
                report_codes.BOOTH_CONFIG_DISTRIBUTION_STARTED,
                {}
            )]
        )

@mock.patch("pcs.lib.booth.sync.run_and_raise")
@mock.patch("pcs.lib.booth.config_structure.get_authfile")
@mock.patch("pcs.lib.booth.config_parser.parse")
@mock.patch("pcs.lib.booth.config_files.read_configs")
@mock.patch("pcs.lib.booth.config_files.read_authfile")
class SendAllConfigToNodeTest(TestCase):
    def setUp(self):
        self.mock_com = "communicator"
        self.mock_reporter = MockLibraryReportProcessor()
        self.node = RequestTarget("node")
        self.file_list = [
            {
                "name": "name1.conf",
                "data": "config1",
                "is_authfile": False
            },
            {
                "name": "file1.key",
                "data": to_b64("some key"),
                "is_authfile": True
            },
            {
                "name": "name2.conf",
                "data": "config2",
                "is_authfile": False
            },
            {
                "name": "file2.key",
                "data": to_b64("another key"),
                "is_authfile": True
            }
        ]

    @staticmethod
    def mock_parse_fn(config_content):
        if config_content not in ["config1", "config2"]:
            raise AssertionError(
                "unexpected input {0}".format(config_content)
            )
        return config_content

    @staticmethod
    def mock_authfile_fn(parsed_config):
        _data = {
            "config1": "/path/to/file1.key",
            "config2": "/path/to/file2.key"
        }
        if parsed_config not in _data:
            raise AssertionError(
                "unexpected input {0}".format(parsed_config)
            )
        return _data[parsed_config]

    @staticmethod
    def mock_read_authfile_fn(_, authfile_path):
        _data = {
            "/path/to/file1.key": "some key".encode("utf-8"),
            "/path/to/file2.key": "another key".encode("utf-8"),
        }
        if authfile_path not in _data:
            raise AssertionError(
                "unexpected input {0}".format(authfile_path)
            )
        return _data[authfile_path]

    def test_success(
        self, mock_read_authfile, mock_read_configs, mock_parse, mock_authfile,
        mock_run_com,
    ):
        mock_parse.side_effect = self.mock_parse_fn
        mock_authfile.side_effect = self.mock_authfile_fn
        mock_read_authfile.side_effect = self.mock_read_authfile_fn
        mock_read_configs.return_value = {
            "name1.conf": "config1",
            "name2.conf": "config2"
        }
        mock_run_com.return_value = [(
            self.node,
            {
                "existing": [],
                "failed": {},
                "saved": ["name1.conf", "file1.key", "name2.conf", "file2.key"]
            }
        )]
        lib.send_all_config_to_node(
            self.mock_com, self.mock_reporter, self.node
        )
        self.assertEqual(2, mock_parse.call_count)
        mock_parse.assert_has_calls([
            mock.call("config1"), mock.call("config2")
        ])
        self.assertEqual(2, mock_authfile.call_count)
        mock_authfile.assert_has_calls([
            mock.call("config1"), mock.call("config2")
        ])
        self.assertEqual(2, mock_read_authfile.call_count)
        mock_read_authfile.assert_has_calls([
            mock.call(self.mock_reporter, "/path/to/file1.key"),
            mock.call(self.mock_reporter, "/path/to/file2.key")
        ])
        mock_read_configs.assert_called_once_with(self.mock_reporter, False)
        communicator, com_cmd = mock_run_com.call_args[0]
        self.assertEqual(self.mock_com, communicator)
        self.assertEqual(self.file_list, com_cmd._file_list)
        self.assertFalse(com_cmd._rewrite_existing)
        assert_report_item_list_equal(
            self.mock_reporter.report_item_list,
            [
                (
                    Severities.INFO,
                    report_codes.BOOTH_CONFIG_DISTRIBUTION_STARTED,
                    {}
                ),
                (
                    Severities.INFO,
                    report_codes.BOOTH_CONFIG_ACCEPTED_BY_NODE,
                    {
                        "node": self.node.label,
                        "name_list": [
                            "name1.conf", "file1.key", "name2.conf", "file2.key"
                        ]
                    }
                )
            ]
        )

    def test_do_not_rewrite_existing(
        self, mock_read_authfile, mock_read_configs, mock_parse, mock_authfile,
        mock_run_com,
    ):
        mock_parse.side_effect = self.mock_parse_fn
        mock_authfile.side_effect = self.mock_authfile_fn
        mock_read_authfile.side_effect = self.mock_read_authfile_fn
        mock_read_configs.return_value = {
            "name1.conf": "config1",
            "name2.conf": "config2"
        }
        mock_run_com.return_value = [(
            self.node,
            {
                "existing": ["name1.conf", "file1.key"],
                "failed": {},
                "saved": ["name2.conf", "file2.key"]
            }
        )]
        assert_raise_library_error(
            lambda: lib.send_all_config_to_node(
                self.mock_com, self.mock_reporter, self.node
            ),
            (
                Severities.ERROR,
                report_codes.FILE_ALREADY_EXISTS,
                {
                    "file_role": None,
                    "file_path": "name1.conf",
                    "node": self.node.label
                },
                report_codes.FORCE_FILE_OVERWRITE
            ),
            (
                Severities.ERROR,
                report_codes.FILE_ALREADY_EXISTS,
                {
                    "file_role": None,
                    "file_path": "file1.key",
                    "node": self.node.label
                },
                report_codes.FORCE_FILE_OVERWRITE
            )
        )
        self.assertEqual(2, mock_parse.call_count)
        mock_parse.assert_has_calls([
            mock.call("config1"), mock.call("config2")
        ])
        self.assertEqual(2, mock_authfile.call_count)
        mock_authfile.assert_has_calls([
            mock.call("config1"), mock.call("config2")
        ])
        self.assertEqual(2, mock_read_authfile.call_count)
        mock_read_authfile.assert_has_calls([
            mock.call(self.mock_reporter, "/path/to/file1.key"),
            mock.call(self.mock_reporter, "/path/to/file2.key")
        ])
        mock_read_configs.assert_called_once_with(self.mock_reporter, False)
        communicator, com_cmd = mock_run_com.call_args[0]
        self.assertEqual(self.mock_com, communicator)
        self.assertEqual(self.file_list, com_cmd._file_list)
        self.assertFalse(com_cmd._rewrite_existing)
        assert_report_item_list_equal(
            self.mock_reporter.report_item_list,
            [
                (
                    Severities.INFO,
                    report_codes.BOOTH_CONFIG_DISTRIBUTION_STARTED,
                    {}
                ),
                (
                    Severities.ERROR,
                    report_codes.FILE_ALREADY_EXISTS,
                    {
                        "file_role": None,
                        "file_path": "name1.conf",
                        "node": self.node.label
                    },
                    report_codes.FORCE_FILE_OVERWRITE
                ),
                (
                    Severities.ERROR,
                    report_codes.FILE_ALREADY_EXISTS,
                    {
                        "file_role": None,
                        "file_path": "file1.key",
                        "node": self.node.label
                    },
                    report_codes.FORCE_FILE_OVERWRITE
                )
            ]
        )

    def test_rewrite_existing(
        self, mock_read_authfile, mock_read_configs, mock_parse, mock_authfile,
        mock_run_com,
    ):
        mock_parse.side_effect = self.mock_parse_fn
        mock_authfile.side_effect = self.mock_authfile_fn
        mock_read_authfile.side_effect = self.mock_read_authfile_fn
        mock_read_configs.return_value = {
            "name1.conf": "config1",
            "name2.conf": "config2"
        }
        mock_run_com.return_value = [(
            self.node,
            {
                "existing": ["name1.conf", "file1.key"],
                "failed": {},
                "saved": ["name2.conf", "file2.key"]
            }
        )]
        lib.send_all_config_to_node(
            self.mock_com,
            self.mock_reporter,
            self.node,
            rewrite_existing=True
        )
        mock_read_configs.assert_called_once_with(self.mock_reporter, False)
        self.assertEqual(2, mock_parse.call_count)
        mock_parse.assert_has_calls([
            mock.call("config1"), mock.call("config2")
        ])
        self.assertEqual(2, mock_authfile.call_count)
        mock_authfile.assert_has_calls([
            mock.call("config1"), mock.call("config2")
        ])
        self.assertEqual(2, mock_read_authfile.call_count)
        mock_read_authfile.assert_has_calls([
            mock.call(self.mock_reporter, "/path/to/file1.key"),
            mock.call(self.mock_reporter, "/path/to/file2.key")
        ])
        communicator, com_cmd = mock_run_com.call_args[0]
        self.assertEqual(self.mock_com, communicator)
        self.assertEqual(self.file_list, com_cmd._file_list)
        self.assertTrue(com_cmd._rewrite_existing)
        assert_report_item_list_equal(
            self.mock_reporter.report_item_list,
            [
                (
                    Severities.INFO,
                    report_codes.BOOTH_CONFIG_DISTRIBUTION_STARTED,
                    {}
                ),
                (
                    Severities.WARNING,
                    report_codes.FILE_ALREADY_EXISTS,
                    {
                        "file_role": None,
                        "file_path": "name1.conf",
                        "node": self.node.label
                    }
                ),
                (
                    Severities.WARNING,
                    report_codes.FILE_ALREADY_EXISTS,
                    {
                        "file_role": None,
                        "file_path": "file1.key",
                        "node": self.node.label
                    }
                ),
                (
                    Severities.INFO,
                    report_codes.BOOTH_CONFIG_ACCEPTED_BY_NODE,
                    {
                        "node": self.node.label,
                        "name_list": ["name2.conf", "file2.key"]
                    }
                )
            ]
        )

    def test_write_failure(
        self, mock_read_authfile, mock_read_configs, mock_parse, mock_authfile,
        mock_run_com,
    ):
        mock_parse.side_effect = self.mock_parse_fn
        mock_authfile.side_effect = self.mock_authfile_fn
        mock_read_authfile.side_effect = self.mock_read_authfile_fn
        mock_read_configs.return_value = {
            "name1.conf": "config1",
            "name2.conf": "config2"
        }
        mock_run_com.return_value = [(
            self.node,
            {
                "existing": [],
                "failed": {
                    "name1.conf": "Error message",
                    "file1.key": "Another error message"
                },
                "saved": ["name2.conf", "file2.key"]
            }
        )]
        assert_raise_library_error(
            lambda: lib.send_all_config_to_node(
                self.mock_com, self.mock_reporter, self.node
            ),
            (
                Severities.ERROR,
                report_codes.BOOTH_CONFIG_DISTRIBUTION_NODE_ERROR,
                {
                    "node": self.node.label,
                    "name": "name1.conf",
                    "reason": "Error message"
                }
            ),
            (
                Severities.ERROR,
                report_codes.BOOTH_CONFIG_DISTRIBUTION_NODE_ERROR,
                {
                    "node": self.node.label,
                    "name": "file1.key",
                    "reason": "Another error message"
                }
            )
        )
        self.assertEqual(2, mock_parse.call_count)
        mock_parse.assert_has_calls([
            mock.call("config1"), mock.call("config2")
        ])
        self.assertEqual(2, mock_authfile.call_count)
        mock_authfile.assert_has_calls([
            mock.call("config1"), mock.call("config2")
        ])
        self.assertEqual(2, mock_read_authfile.call_count)
        mock_read_authfile.assert_has_calls([
            mock.call(self.mock_reporter, "/path/to/file1.key"),
            mock.call(self.mock_reporter, "/path/to/file2.key")
        ])
        mock_read_configs.assert_called_once_with(self.mock_reporter, False)
        communicator, com_cmd = mock_run_com.call_args[0]
        self.assertEqual(self.mock_com, communicator)
        self.assertEqual(self.file_list, com_cmd._file_list)
        self.assertFalse(com_cmd._rewrite_existing)
        assert_report_item_list_equal(
            self.mock_reporter.report_item_list,
            [
                (
                    Severities.INFO,
                    report_codes.BOOTH_CONFIG_DISTRIBUTION_STARTED,
                    {}
                ),
                (
                    Severities.ERROR,
                    report_codes.BOOTH_CONFIG_DISTRIBUTION_NODE_ERROR,
                    {
                        "node": self.node.label,
                        "name": "name1.conf",
                        "reason": "Error message"
                    }
                ),
                (
                    Severities.ERROR,
                    report_codes.BOOTH_CONFIG_DISTRIBUTION_NODE_ERROR,
                    {
                        "node": self.node.label,
                        "name": "file1.key",
                        "reason": "Another error message"
                    }
                )
            ]
        )

    @skip("TODO: rewrite for pcs.lib.communication.booth.BoothSaveFiles")
    def test_communication_failure(
        self, mock_read_authfile, mock_read_configs, mock_parse, mock_authfile
    ):
        mock_parse.side_effect = self.mock_parse_fn
        mock_authfile.side_effect = self.mock_authfile_fn
        mock_read_authfile.side_effect = self.mock_read_authfile_fn
        mock_read_configs.return_value = {
            "name1.conf": "config1",
            "name2.conf": "config2"
        }
        self.mock_communicator.call_node.side_effect = NodeConnectionException(
            self.node.label, "command", "reason"
        )
        assert_raise_library_error(
            lambda: lib.send_all_config_to_node(
                self.mock_communicator, self.mock_reporter, self.node
            ),
            (
                Severities.ERROR,
                report_codes.NODE_COMMUNICATION_ERROR_UNABLE_TO_CONNECT,
                {
                    "node": self.node.label,
                    "command": "command",
                    "reason": "reason"
                }
            )
        )
        self.assertEqual(2, mock_parse.call_count)
        mock_parse.assert_has_calls([
            mock.call("config1"), mock.call("config2")
        ])
        self.assertEqual(2, mock_authfile.call_count)
        mock_authfile.assert_has_calls([
            mock.call("config1"), mock.call("config2")
        ])
        self.assertEqual(2, mock_read_authfile.call_count)
        mock_read_authfile.assert_has_calls([
            mock.call(self.mock_reporter, "/path/to/file1.key"),
            mock.call(self.mock_reporter, "/path/to/file2.key")
        ])
        mock_read_configs.assert_called_once_with(self.mock_reporter, False)
        self.assertEqual(1, self.mock_communicator.call_node.call_count)
        self.assertEqual(
            self.node, self.mock_communicator.call_node.call_args[0][0]
        )
        self.assertEqual(
            "remote/booth_save_files",
            self.mock_communicator.call_node.call_args[0][1]
        )
        data = url_decode(self.mock_communicator.call_node.call_args[0][2])
        self.assertFalse("rewrite_existing" in data)
        self.assertTrue("data_json" in data)
        self.assertEqual(
            [
                {
                    "name": "name1.conf",
                    "data": "config1",
                    "is_authfile": False
                },
                {
                    "name": "file1.key",
                    "data": to_b64("some key"),
                    "is_authfile": True
                },
                {
                    "name": "name2.conf",
                    "data": "config2",
                    "is_authfile": False
                },
                {
                    "name": "file2.key",
                    "data": to_b64("another key"),
                    "is_authfile": True
                }
            ],
            json.loads(data["data_json"][0])
        )

    @skip("TODO: rewrite for pcs.lib.communication.booth.BoothSaveFiles")
    def test_wrong_response_format(
        self, mock_read_authfile, mock_read_configs, mock_parse, mock_authfile
    ):
        mock_parse.side_effect = self.mock_parse_fn
        mock_authfile.side_effect = self.mock_authfile_fn
        mock_read_authfile.side_effect = self.mock_read_authfile_fn
        mock_read_configs.return_value = {
            "name1.conf": "config1",
            "name2.conf": "config2"
        }
        self.mock_communicator.call_node.return_value = """
            {
                "existing_files": [],
                "failed": {
                    "name1.conf": "Error message",
                    "file1.key": "Another error message"
                },
                "saved": ["name2.conf", "file2.key"]
            }
        """
        assert_raise_library_error(
            lambda: lib.send_all_config_to_node(
                self.mock_communicator, self.mock_reporter, self.node
            ),
            (
                Severities.ERROR,
                report_codes.INVALID_RESPONSE_FORMAT,
                {"node": self.node.label}
            )
        )
        self.assertEqual(2, mock_parse.call_count)
        mock_parse.assert_has_calls([
            mock.call("config1"), mock.call("config2")
        ])
        self.assertEqual(2, mock_authfile.call_count)
        mock_authfile.assert_has_calls([
            mock.call("config1"), mock.call("config2")
        ])
        self.assertEqual(2, mock_read_authfile.call_count)
        mock_read_authfile.assert_has_calls([
            mock.call(self.mock_reporter, "/path/to/file1.key"),
            mock.call(self.mock_reporter, "/path/to/file2.key")
        ])
        mock_read_configs.assert_called_once_with(self.mock_reporter, False)
        self.assertEqual(1, self.mock_communicator.call_node.call_count)
        self.assertEqual(
            self.node, self.mock_communicator.call_node.call_args[0][0]
        )
        self.assertEqual(
            "remote/booth_save_files",
            self.mock_communicator.call_node.call_args[0][1]
        )
        data = url_decode(self.mock_communicator.call_node.call_args[0][2])
        self.assertFalse("rewrite_existing" in data)
        self.assertTrue("data_json" in data)
        self.assertEqual(
            [
                {
                    "name": "name1.conf",
                    "data": "config1",
                    "is_authfile": False
                },
                {
                    "name": "file1.key",
                    "data": to_b64("some key"),
                    "is_authfile": True
                },
                {
                    "name": "name2.conf",
                    "data": "config2",
                    "is_authfile": False
                },
                {
                    "name": "file2.key",
                    "data": to_b64("another key"),
                    "is_authfile": True
                }
            ],
            json.loads(data["data_json"][0])
        )

    @skip("TODO: rewrite for pcs.lib.communication.booth.BoothSaveFiles")
    def test_response_not_json(
        self, mock_read_authfile, mock_read_configs, mock_parse, mock_authfile
    ):
        mock_parse.side_effect = self.mock_parse_fn
        mock_authfile.side_effect = self.mock_authfile_fn
        mock_read_authfile.side_effect = self.mock_read_authfile_fn
        mock_read_configs.return_value = {
            "name1.conf": "config1",
            "name2.conf": "config2"
        }
        self.mock_communicator.call_node.return_value = "not json"
        assert_raise_library_error(
            lambda: lib.send_all_config_to_node(
                self.mock_communicator, self.mock_reporter, self.node
            ),
            (
                Severities.ERROR,
                report_codes.INVALID_RESPONSE_FORMAT,
                {"node": self.node.label}
            )
        )
        self.assertEqual(2, mock_parse.call_count)
        mock_parse.assert_has_calls([
            mock.call("config1"), mock.call("config2")
        ])
        self.assertEqual(2, mock_authfile.call_count)
        mock_authfile.assert_has_calls([
            mock.call("config1"), mock.call("config2")
        ])
        self.assertEqual(2, mock_read_authfile.call_count)
        mock_read_authfile.assert_has_calls([
            mock.call(self.mock_reporter, "/path/to/file1.key"),
            mock.call(self.mock_reporter, "/path/to/file2.key")
        ])
        mock_read_configs.assert_called_once_with(self.mock_reporter, False)
        self.assertEqual(1, self.mock_communicator.call_node.call_count)
        self.assertEqual(
            self.node, self.mock_communicator.call_node.call_args[0][0]
        )
        self.assertEqual(
            "remote/booth_save_files",
            self.mock_communicator.call_node.call_args[0][1]
        )
        data = url_decode(self.mock_communicator.call_node.call_args[0][2])
        self.assertFalse("rewrite_existing" in data)
        self.assertTrue("data_json" in data)
        self.assertEqual(
            [
                {
                    "name": "name1.conf",
                    "data": "config1",
                    "is_authfile": False
                },
                {
                    "name": "file1.key",
                    "data": to_b64("some key"),
                    "is_authfile": True
                },
                {
                    "name": "name2.conf",
                    "data": "config2",
                    "is_authfile": False
                },
                {
                    "name": "file2.key",
                    "data": to_b64("another key"),
                    "is_authfile": True
                }
            ],
            json.loads(data["data_json"][0])
        )

    def test_configs_without_authfiles(
        self, mock_read_authfile, mock_read_configs, mock_parse, mock_authfile,
        mock_run_com
    ):
        def mock_authfile_fn(parsed_config):
            if parsed_config == "config1":
                return None
            elif parsed_config == "config2":
                return "/path/to/file2.key"
            else:
                raise AssertionError(
                    "unexpected input: {0}".format(parsed_config)
                )

        mock_parse.side_effect = self.mock_parse_fn
        mock_authfile.side_effect = mock_authfile_fn
        mock_read_authfile.return_value = "another key".encode("utf-8")
        mock_read_configs.return_value = {
            "name1.conf": "config1",
            "name2.conf": "config2"
        }
        mock_run_com.return_value = [(
            self.node,
            {
                "existing": [],
                "failed": {},
                "saved": ["name1.conf", "name2.conf", "file2.key"]
            }
        )]
        lib.send_all_config_to_node(
            self.mock_com, self.mock_reporter, self.node
        )
        self.assertEqual(2, mock_parse.call_count)
        mock_parse.assert_has_calls([
            mock.call("config1"), mock.call("config2")
        ])
        self.assertEqual(2, mock_authfile.call_count)
        mock_authfile.assert_has_calls([
            mock.call("config1"), mock.call("config2")
        ])
        mock_read_authfile.assert_called_once_with(
            self.mock_reporter, "/path/to/file2.key"
        )
        mock_read_configs.assert_called_once_with(self.mock_reporter, False)
        expected_file_list = [
            {
                "name": "name1.conf",
                "data": "config1",
                "is_authfile": False
            },
            {
                "name": "name2.conf",
                "data": "config2",
                "is_authfile": False
            },
            {
                "name": "file2.key",
                "data": to_b64("another key"),
                "is_authfile": True
            }
        ]
        communicator, com_cmd = mock_run_com.call_args[0]
        self.assertEqual(self.mock_com, communicator)
        self.assertEqual(expected_file_list, com_cmd._file_list)
        self.assertFalse(com_cmd._rewrite_existing)
        assert_report_item_list_equal(
            self.mock_reporter.report_item_list,
            [
                (
                    Severities.INFO,
                    report_codes.BOOTH_CONFIG_DISTRIBUTION_STARTED,
                    {}
                ),
                (
                    Severities.INFO,
                    report_codes.BOOTH_CONFIG_ACCEPTED_BY_NODE,
                    {
                        "node": self.node.label,
                        "name_list": ["name1.conf", "name2.conf", "file2.key"]
                    }
                )
            ]
        )

    def test_unable_to_parse_config(
        self, mock_read_authfile, mock_read_configs, mock_parse, mock_authfile,
        mock_run_com,
    ):
        def mock_parse_fn(config_data):
            if config_data == "config1":
                raise LibraryError()
            elif config_data == "config2":
                return "config2"
            else:
                raise AssertionError(
                    "unexpected input: {0}".format(config_data)
                )

        mock_parse.side_effect = mock_parse_fn
        mock_authfile.return_value = "/path/to/file2.key"
        mock_read_authfile.return_value = "another key".encode("utf-8")
        mock_read_configs.return_value = {
            "name1.conf": "config1",
            "name2.conf": "config2"
        }
        mock_run_com.return_value = [(
            self.node,
            {
                "existing": [],
                "failed": {},
                "saved": ["name2.conf", "file2.key"]
            }
        )]
        lib.send_all_config_to_node(
            self.mock_com, self.mock_reporter, self.node
        )
        self.assertEqual(2, mock_parse.call_count)
        mock_parse.assert_has_calls([
            mock.call("config1"), mock.call("config2")
        ])
        mock_authfile.assert_called_once_with("config2")
        mock_read_authfile.assert_called_once_with(
            self.mock_reporter, "/path/to/file2.key"
        )
        mock_read_configs.assert_called_once_with(self.mock_reporter, False)
        expected_file_list = [
            {
                "name": "name2.conf",
                "data": "config2",
                "is_authfile": False
            },
            {
                "name": "file2.key",
                "data": to_b64("another key"),
                "is_authfile": True
            }
        ]
        communicator, com_cmd = mock_run_com.call_args[0]
        self.assertEqual(self.mock_com, communicator)
        self.assertEqual(expected_file_list, com_cmd._file_list)
        self.assertFalse(com_cmd._rewrite_existing)
        assert_report_item_list_equal(
            self.mock_reporter.report_item_list,
            [
                (
                    Severities.INFO,
                    report_codes.BOOTH_CONFIG_DISTRIBUTION_STARTED,
                    {}
                ),
                (
                    Severities.WARNING,
                    report_codes.BOOTH_SKIPPING_CONFIG,
                    {
                        "config_file": "name1.conf"
                    }
                ),
                (
                    Severities.INFO,
                    report_codes.BOOTH_CONFIG_ACCEPTED_BY_NODE,
                    {
                        "node": self.node.label,
                        "name_list": ["name2.conf", "file2.key"]
                    }
                )
            ]
        )



from __future__ import (
    absolute_import,
    division,
    print_function,
)

import os.path
from collections import namedtuple

from pcs.cli.common import console_report
from pcs.common import report_codes


def report_missing(file_role, file_path):
    console_report.error(
        "{0} '{1}' does not exist".format(file_role, file_path)
    )

def is_missing_report(report, file_role_code):
    return (
        report.code == report_codes.FILE_DOES_NOT_EXIST
        and
        report.info["file_role"] == file_role_code
    )

def process_no_existing_file_expectation(file_role, env_file, file_path):
    if(
        env_file["no_existing_file_expected"]
        and
        os.path.exists(file_path)
    ):
        msg = "{0} {1} already exists".format(file_role, file_path)
        if not env_file["can_overwrite_existing_file"]:
            raise console_report.error(
                "{0}, use --force to override".format(msg)
            )
        console_report.warn(msg)

def write(env_file, file_path):
    try:
        f = open(file_path, "wb" if env_file.get("is_binary", False) else "w")
        f.write(env_file["content"])
        f.close()
    except EnvironmentError as e:
        raise console_report.error(
            "Unable to write {0}: {1}".format(file_path, e.strerror)
        )

def read(path, is_binary=False):
    try:
        mode = "rb" if is_binary else "r"
        return {
            "content": open(path, mode).read() if os.path.isfile(path) else None
        }
    except EnvironmentError as e:
        raise console_report.error(
            "Unable to read {0}: {1}".format(path, e.strerror)
        )

MissingFileCandidateInfo = namedtuple(
    "MissingFileCandidateInfo",
    "code desc path"
)

def evaluate_for_missing_files(exception, file_info_list):
    """
    list of MissingFileCandidateInfo file_info_list contains the info for files
        that can be missing
    """
    for report in exception.args:
        for file_info in file_info_list:
            if is_missing_report(report, file_info.code):
                report_missing(file_info.desc, file_info.path)
                exception.sign_processed(report)



from __future__ import (
    absolute_import,
    division,
    print_function,
)

from lxml import etree
import shutil

from pcs.test.tools.assertions import AssertPcsMixin
from pcs.test.tools.cib import get_assert_pcs_effect_mixin
from pcs.test.tools.pcs_unittest import TestCase
from pcs.test.tools.misc import (
    get_test_resource as rc,
    outdent,
    skip_unless_pacemaker_supports_bundle,
)
from pcs.test.tools.pcs_runner import PcsRunner


class BundleCreateCommon(
    TestCase,
    get_assert_pcs_effect_mixin(
        lambda cib: etree.tostring(
            # pylint:disable=undefined-variable
            etree.parse(cib).findall(".//resources")[0]
        )
    )
):
    temp_cib = rc("temp-cib.xml")

    def setUp(self):
        shutil.copy(self.empty_cib, self.temp_cib)
        self.pcs_runner = PcsRunner(self.temp_cib)


@skip_unless_pacemaker_supports_bundle
class BundleCreateUpgradeCib(BundleCreateCommon):
    empty_cib = rc("cib-empty.xml")

    def test_success(self):
        self.assert_effect(
            "resource bundle create B1 container docker image=pcs:test",
            """
                <resources>
                    <bundle id="B1">
                        <docker image="pcs:test" />
                    </bundle>
                </resources>
            """,
            "CIB has been upgraded to the latest schema version.\n"
        )


@skip_unless_pacemaker_supports_bundle
class BundleCreate(BundleCreateCommon):
    empty_cib = rc("cib-empty-2.8.xml")

    def test_minimal(self):
        self.assert_effect(
            "resource bundle create B1 container docker image=pcs:test",
            """
                <resources>
                    <bundle id="B1">
                        <docker image="pcs:test" />
                    </bundle>
                </resources>
            """
        )

    def test_all_options(self):
        self.assert_effect(
            """
                resource bundle create B1
                container docker replicas=4 replicas-per-host=2
                    run-command=/bin/true
                port-map port=1001
                meta target-role=Stopped
                network control-port=12345 host-interface=eth0 host-netmask=24
                port-map id=B1-port-map-1001 internal-port=2002 port=2000
                port-map range=3000-3300
                storage-map source-dir=/tmp/docker1a target-dir=/tmp/docker1b
                network ip-range-start=192.168.100.200
                storage-map id=B1-storage-map source-dir=/tmp/docker2a
                    target-dir=/tmp/docker2b
                container image=pcs:test masters=0
                meta is-managed=false
                storage-map source-dir-root=/tmp/docker3a
                    target-dir=/tmp/docker3b
                storage-map id=B1-port-map-1001-1 source-dir-root=/tmp/docker4a
                    target-dir=/tmp/docker4b
                container network=extra_network_settings options=extra_options
            """,
            """
                <resources>
                    <bundle id="B1">
                        <docker
                            image="pcs:test"
                            masters="0"
                            network="extra_network_settings"
                            options="extra_options"
                            replicas="4"
                            replicas-per-host="2"
                            run-command="/bin/true"
                        />
                        <network
                            control-port="12345"
                            host-interface="eth0"
                            host-netmask="24"
                            ip-range-start="192.168.100.200"
                        >
                            <port-mapping id="B1-port-map-1001-2" port="1001" />
                            <port-mapping
                                id="B1-port-map-1001"
                                internal-port="2002"
                                port="2000"
                            />
                            <port-mapping
                                id="B1-port-map-3000-3300"
                                range="3000-3300"
                            />
                        </network>
                        <storage>
                            <storage-mapping
                                id="B1-storage-map-1"
                                source-dir="/tmp/docker1a"
                                target-dir="/tmp/docker1b"
                            />
                            <storage-mapping
                                id="B1-storage-map"
                                source-dir="/tmp/docker2a"
                                target-dir="/tmp/docker2b"
                            />
                            <storage-mapping
                                id="B1-storage-map-2"
                                source-dir-root="/tmp/docker3a"
                                target-dir="/tmp/docker3b"
                            />
                            <storage-mapping
                                id="B1-port-map-1001-1"
                                source-dir-root="/tmp/docker4a"
                                target-dir="/tmp/docker4b"
                            />
                        </storage>
                        <meta_attributes id="B1-meta_attributes">
                            <nvpair
                                id="B1-meta_attributes-is-managed"
                                name="is-managed"
                                value="false"
                            />
                            <nvpair
                                id="B1-meta_attributes-target-role"
                                name="target-role"
                                value="Stopped"
                            />
                        </meta_attributes>
                    </bundle>
                </resources>
            """
        )

    def test_fail_when_missing_args_1(self):
        self.assert_pcs_fail_regardless_of_force(
            "resource bundle",
            stdout_start="\nUsage: pcs resource bundle ...\n"
        )

    def test_fail_when_missing_args_2(self):
        self.assert_pcs_fail_regardless_of_force(
            "resource bundle create",
            stdout_start="\nUsage: pcs resource bundle create...\n"
        )

    def test_fail_when_missing_container_type(self):
        self.assert_pcs_fail_regardless_of_force(
            "resource bundle create B1",
            "Error: '' is not a valid container type value, use docker\n"
        )

    def test_fail_when_missing_required(self):
        self.assert_pcs_fail_regardless_of_force(
            "resource bundle create B1 container docker",
            "Error: required container option 'image' is missing\n"
        )

    def test_fail_on_unknown_option(self):
        self.assert_pcs_fail(
            """
                resource bundle create B1 container docker image=pcs:test
                extra=option
            """,
            "Error: invalid container option 'extra', allowed options are: "
                "image, masters, network, options, replicas, replicas-per-host,"
                " run-command, use --force to override\n"
        )

    def test_unknown_option_forced(self):
        # Test that pcs allows to specify options it does not know about. This
        # ensures some kind of forward compatibility, so the user will be able
        # to specify new options. However, as of now the option is not
        # supported by pacemaker and so the command fails.
        self.assert_pcs_fail(
            """
                resource bundle create B1 container docker image=pcs:test
                extra=option --force
            """
            ,
            stdout_start="Error: Unable to update cib\n"
        )

    def test_more_errors(self):
        self.assert_pcs_fail_regardless_of_force(
            "resource bundle create B#1 container docker replicas=x",
            outdent(
                """\
                Error: invalid bundle name 'B#1', '#' is not a valid character for a bundle name
                Error: required container option 'image' is missing
                Error: 'x' is not a valid replicas value, use a positive integer
                """
            )
        )

    def assert_no_options(self, keyword):
        self.assert_pcs_fail_regardless_of_force(
            "resource bundle create B {0}".format(keyword),
            "Error: No {0} options specified\n".format(keyword),
        )

    def test_empty_container(self):
        self.assert_no_options("container")

    def test_empty_network(self):
        self.assert_no_options("network")

    def test_empty_storage_map(self):
        self.assert_no_options("storage-map")

    def test_empty_port_map(self):
        self.assert_no_options("port-map")

    def test_empty_meta(self):
        self.assert_no_options("meta")


@skip_unless_pacemaker_supports_bundle
class BundleUpdate(BundleCreateCommon):
    empty_cib = rc("cib-empty-2.8.xml")

    def fixture_bundle(self, name):
        self.assert_pcs_success(
            "resource bundle create {0} container docker image=pcs:test".format(
                name
            )
        )

    def fixture_bundle_complex(self, name):
        self.assert_pcs_success(
            ("""
                resource bundle create {0}
                container docker image=pcs:test replicas=4 masters=2
                network control-port=12345 host-interface=eth0 host-netmask=24
                port-map internal-port=1000 port=2000
                port-map internal-port=1001 port=2001
                port-map internal-port=1002 port=2002
                storage-map source-dir=/tmp/docker1a target-dir=/tmp/docker1b
                storage-map source-dir=/tmp/docker2a target-dir=/tmp/docker2b
                storage-map source-dir=/tmp/docker3a target-dir=/tmp/docker3b
                meta priority=15 resource-stickiness=100 is-managed=false
            """).format(name)
        )

    def test_fail_when_missing_args_1(self):
        self.assert_pcs_fail_regardless_of_force(
            "resource bundle update",
            stdout_start="\nUsage: pcs resource bundle update...\n"
        )

    def test_fail_when_missing_args_2(self):
        self.fixture_bundle("B")
        self.assert_pcs_fail_regardless_of_force(
            "resource bundle update B port-map",
            "Error: No port-map options specified\n"
        )

    def test_fail_when_missing_args_3(self):
        self.fixture_bundle("B")
        self.assert_pcs_fail_regardless_of_force(
            "resource bundle update B storage-map remove",
            "Error: When using 'storage-map' you must specify either 'add' and "
                "options or 'remove' and id(s)\n"
        )

    def test_bad_id(self):
        self.fixture_bundle("B")
        self.assert_pcs_fail_regardless_of_force(
            "resource bundle update B1 container image=test",
            "Error: bundle 'B1' does not exist\n"
        )

    def test_success(self):
        self.fixture_bundle_complex("B")
        self.assert_effect(
            """
                resource bundle update B
                container masters= replicas=6 replicas-per-host=2
                network control-port= host-interface=eth1
                    ip-range-start=192.168.100.200
                port-map remove B-port-map-2000 B-port-map-2002
                port-map add internal-port=1003 port=2003
                storage-map remove B-storage-map B-storage-map-2
                storage-map add source-dir=/tmp/docker4a target-dir=/tmp/docker4b
                meta priority=10 is-managed= target-role=Stopped
            """,
            """
                <resources>
                    <bundle id="B">
                        <docker
                            image="pcs:test"
                            replicas="6"
                            replicas-per-host="2"
                        />
                        <network
                            host-interface="eth1"
                            host-netmask="24"
                            ip-range-start="192.168.100.200"
                        >
                            <port-mapping
                                id="B-port-map-2001"
                                internal-port="1001"
                                port="2001"
                            />
                            <port-mapping
                                id="B-port-map-2003"
                                internal-port="1003"
                                port="2003"
                            />
                        </network>
                        <storage>
                            <storage-mapping
                                id="B-storage-map-1"
                                source-dir="/tmp/docker2a"
                                target-dir="/tmp/docker2b"
                            />
                            <storage-mapping
                                id="B-storage-map"
                                source-dir="/tmp/docker4a"
                                target-dir="/tmp/docker4b"
                            />
                        </storage>
                        <meta_attributes id="B-meta_attributes">
                            <nvpair id="B-meta_attributes-priority"
                                name="priority" value="10" />
                            <nvpair id="B-meta_attributes-resource-stickiness"
                                name="resource-stickiness" value="100" />
                            <nvpair id="B-meta_attributes-target-role"
                                name="target-role" value="Stopped" />
                        </meta_attributes>
                    </bundle>
                </resources>
            """
        )

    def test_force_unknown_option(self):
        self.fixture_bundle("B")

        self.assert_pcs_fail(
            "resource bundle update B container extra=option",
            "Error: invalid container option 'extra', allowed options are: "
                "image, masters, network, options, replicas, replicas-per-host,"
                " run-command, use --force to override\n"
        )
        # Test that pcs allows to specify options it does not know about. This
        # ensures some kind of forward compatibility, so the user will be able
        # to specify new options. However, as of now the option is not
        # supported by pacemaker and so the command fails.
        self.assert_pcs_fail(
            "resource bundle update B container extra=option --force",
            stdout_start="Error: Unable to update cib\n"
        )

        # no force needed when removing an unknown option
        self.assert_effect(
            "resource bundle update B container extra=",
            """
                <resources>
                    <bundle id="B">
                        <docker image="pcs:test" />
                    </bundle>
                </resources>
            """
        )

    def assert_no_options(self, keyword):
        self.fixture_bundle("B")
        self.assert_pcs_fail_regardless_of_force(
            "resource bundle update B {0}".format(keyword),
            "Error: No {0} options specified\n".format(keyword),
        )

    def test_empty_container(self):
        self.assert_no_options("container")

    def test_empty_network(self):
        self.assert_no_options("network")

    def test_empty_storage_map(self):
        self.assert_no_options("storage-map")

    def test_empty_port_map(self):
        self.assert_no_options("port-map")

    def test_empty_meta(self):
        self.assert_no_options("meta")


@skip_unless_pacemaker_supports_bundle
class BundleShow(TestCase, AssertPcsMixin):
    empty_cib = rc("cib-empty-2.8.xml")
    temp_cib = rc("temp-cib.xml")

    def setUp(self):
        shutil.copy(self.empty_cib, self.temp_cib)
        self.pcs_runner = PcsRunner(self.temp_cib)

    def test_minimal(self):
        self.assert_pcs_success(
            "resource bundle create B1 container docker image=pcs:test"
        )
        self.assert_pcs_success("resource show B1", outdent(
            """\
             Bundle: B1
              Docker: image=pcs:test
            """
        ))

    def test_container(self):
        self.assert_pcs_success(
            """
                resource bundle create B1
                container docker image=pcs:test masters=2 replicas=4
                    options='a b c'
            """
        )
        self.assert_pcs_success("resource show B1", outdent(
            """\
             Bundle: B1
              Docker: image=pcs:test masters=2 options="a b c" replicas=4
            """
        ))

    def test_network(self):
        self.assert_pcs_success(
            """
                resource bundle create B1
                container docker image=pcs:test
                network host-interface=eth0 host-netmask=24 control-port=12345
            """
        )
        self.assert_pcs_success("resource show B1", outdent(
            """\
             Bundle: B1
              Docker: image=pcs:test
              Network: control-port=12345 host-interface=eth0 host-netmask=24
            """
        ))

    def test_port_map(self):
        self.assert_pcs_success(
            """
                resource bundle create B1
                container docker image=pcs:test
                port-map id=B1-port-map-1001 internal-port=2002 port=2000
                port-map range=3000-3300
            """
        )
        self.assert_pcs_success("resource show B1", outdent(
            """\
             Bundle: B1
              Docker: image=pcs:test
              Port Mapping:
               internal-port=2002 port=2000 (B1-port-map-1001)
               range=3000-3300 (B1-port-map-3000-3300)
            """
        ))

    def test_storage_map(self):
        self.assert_pcs_success(
            """
                resource bundle create B1
                container docker image=pcs:test
                storage-map source-dir=/tmp/docker1a target-dir=/tmp/docker1b
                storage-map id=my-storage-map source-dir=/tmp/docker2a
                    target-dir=/tmp/docker2b
            """
        )
        self.assert_pcs_success("resource show B1", outdent(
            """\
             Bundle: B1
              Docker: image=pcs:test
              Storage Mapping:
               source-dir=/tmp/docker1a target-dir=/tmp/docker1b (B1-storage-map)
               source-dir=/tmp/docker2a target-dir=/tmp/docker2b (my-storage-map)
            """
        ))

    def test_meta(self):
        self.assert_pcs_success("""
            resource bundle create B1 container docker image=pcs:test
            --disabled
        """)
        self.assert_pcs_success("resource show B1", outdent(
            # pylint:disable=trailing-whitespace
            """\
             Bundle: B1
              Docker: image=pcs:test
              Meta Attrs: target-role=Stopped 
            """
        ))

    def test_resource(self):
        self.assert_pcs_success(
            "resource bundle create B1 container docker image=pcs:test network "
            "control-port=1234"
        )
        self.assert_pcs_success(
            "resource create A ocf:pacemaker:Dummy bundle B1 --no-default-ops"
        )
        self.assert_pcs_success("resource show B1", outdent(
            """\
             Bundle: B1
              Docker: image=pcs:test
              Network: control-port=1234
              Resource: A (class=ocf provider=pacemaker type=Dummy)
               Operations: monitor interval=10 timeout=20 (A-monitor-interval-10)
            """
        ))

    def test_all(self):
        self.assert_pcs_success(
            """
                resource bundle create B1
                container docker image=pcs:test masters=2 replicas=4
                    options='a b c'
                network host-interface=eth0 host-netmask=24 control-port=12345
                port-map id=B1-port-map-1001 internal-port=2002 port=2000
                port-map range=3000-3300
                storage-map source-dir=/tmp/docker1a target-dir=/tmp/docker1b
                storage-map id=my-storage-map source-dir=/tmp/docker2a
                    target-dir=/tmp/docker2b
                meta target-role=Stopped is-managed=false
            """
        )
        self.assert_pcs_success(
            "resource create A ocf:pacemaker:Dummy bundle B1 --no-default-ops"
        )
        self.assert_pcs_success("resource show B1", outdent(
            # pylint:disable=trailing-whitespace
            """\
             Bundle: B1
              Docker: image=pcs:test masters=2 options="a b c" replicas=4
              Network: control-port=12345 host-interface=eth0 host-netmask=24
              Port Mapping:
               internal-port=2002 port=2000 (B1-port-map-1001)
               range=3000-3300 (B1-port-map-3000-3300)
              Storage Mapping:
               source-dir=/tmp/docker1a target-dir=/tmp/docker1b (B1-storage-map)
               source-dir=/tmp/docker2a target-dir=/tmp/docker2b (my-storage-map)
              Meta Attrs: is-managed=false target-role=Stopped 
              Resource: A (class=ocf provider=pacemaker type=Dummy)
               Operations: monitor interval=10 timeout=20 (A-monitor-interval-10)
            """
        ))



from __future__ import (
    absolute_import,
    division,
    print_function,
)

LOG_FILE = "/var/log/pcsd/pcs_snmp_agent.log"
ENTERPRISES_OID = "1.3.6.1.4.1"
PACEMAKER_OID = ENTERPRISES_OID + ".32723"
PCS_OID = PACEMAKER_OID + ".100"
DEFAULT_UPDATE_INTERVAL = 30



from __future__ import (
    absolute_import,
    division,
    print_function,
)

from pcs.test.tools import pcs_unittest as unittest

from pcs.test.tools.assertions import ac

from pcs.lib.corosync import config_parser


class SectionTest(unittest.TestCase):

    def test_empty_section(self):
        section = config_parser.Section("mySection")
        self.assertEqual(section.parent, None)
        self.assertEqual(section.get_root(), section)
        self.assertEqual(section.name, "mySection")
        self.assertEqual(section.get_attributes(), [])
        self.assertEqual(section.get_sections(), [])
        self.assertTrue(section.empty)
        ac(str(section), "")

    def test_is_section_empty(self):
        section = config_parser.Section("mySection")
        self.assertTrue(section.empty)

        section = config_parser.Section("mySection")
        section.add_attribute("name", "value")
        self.assertFalse(section.empty)

        section = config_parser.Section("mySection")
        section.add_section(config_parser.Section("subSection"))
        self.assertFalse(section.empty)

        section = config_parser.Section("mySection")
        section.add_attribute("name", "value")
        section.add_section(config_parser.Section("subSection"))
        self.assertFalse(section.empty)

    def test_attribute_add(self):
        section = config_parser.Section("mySection")

        section.add_attribute("name1", "value1")
        self.assertEqual(
            section.get_attributes(),
            [
                ["name1", "value1"],
            ]
        )

        section.add_attribute("name2", "value2")
        self.assertEqual(
            section.get_attributes(),
            [
                ["name1", "value1"],
                ["name2", "value2"],
            ]
        )

        section.add_attribute("name2", "value2")
        self.assertEqual(
            section.get_attributes(),
            [
                ["name1", "value1"],
                ["name2", "value2"],
                ["name2", "value2"],
            ]
        )

    def test_attribute_get(self):
        section = config_parser.Section("mySection")
        section.add_attribute("name1", "value1")
        section.add_attribute("name2", "value2")
        section.add_attribute("name3", "value3")
        section.add_attribute("name2", "value2a")

        self.assertEqual(
            section.get_attributes(),
            [
                ["name1", "value1"],
                ["name2", "value2"],
                ["name3", "value3"],
                ["name2", "value2a"],
            ]
        )
        self.assertEqual(
            section.get_attributes("name1"),
            [
                ["name1", "value1"],
            ]
        )
        self.assertEqual(
            section.get_attributes("name2"),
            [
                ["name2", "value2"],
                ["name2", "value2a"],
            ]
        )
        self.assertEqual(
            section.get_attributes("nameX"),
            []
        )

    def test_attribute_set(self):
        section = config_parser.Section("mySection")

        section.set_attribute("name1", "value1")
        self.assertEqual(
            section.get_attributes(),
            [
                ["name1", "value1"],
            ]
        )

        section.set_attribute("name1", "value1")
        self.assertEqual(
            section.get_attributes(),
            [
                ["name1", "value1"],
            ]
        )

        section.set_attribute("name1", "value1a")
        self.assertEqual(
            section.get_attributes(),
            [
                ["name1", "value1a"],
            ]
        )

        section.set_attribute("name2", "value2")
        self.assertEqual(
            section.get_attributes(),
            [
                ["name1", "value1a"],
                ["name2", "value2"],
            ]
        )

        section.set_attribute("name1", "value1")
        self.assertEqual(
            section.get_attributes(),
            [
                ["name1", "value1"],
                ["name2", "value2"],
            ]
        )

        section.add_attribute("name3", "value3")
        section.add_attribute("name2", "value2")
        self.assertEqual(
            section.get_attributes(),
            [
                ["name1", "value1"],
                ["name2", "value2"],
                ["name3", "value3"],
                ["name2", "value2"],
            ]
        )
        section.set_attribute("name2", "value2a")
        self.assertEqual(
            section.get_attributes(),
            [
                ["name1", "value1"],
                ["name2", "value2a"],
                ["name3", "value3"],
            ]
        )

        section.add_attribute("name1", "value1")
        section.add_attribute("name1", "value1")
        section.set_attribute("name1", "value1")
        self.assertEqual(
            section.get_attributes(),
            [
                ["name1", "value1"],
                ["name2", "value2a"],
                ["name3", "value3"],
            ]
        )

    def test_attribute_change(self):
        section = config_parser.Section("mySection")
        section.add_attribute("name1", "value1")
        section.add_attribute("name2", "value2")
        section.add_attribute("name3", "value3")
        section.add_attribute("name2", "value2")

        attr = section.get_attributes()[1]
        attr[0] = "name2a"
        attr[1] = "value2a"
        self.assertEqual(
            section.get_attributes(),
            [
                ["name1", "value1"],
                ["name2a", "value2a"],
                ["name3", "value3"],
                ["name2", "value2"],
            ]
        )

    def test_attribute_del(self):
        section = config_parser.Section("mySection")
        section.add_attribute("name1", "value1")
        section.add_attribute("name2", "value2")
        section.add_attribute("name3", "value3")
        section.add_attribute("name2", "value2")

        section.del_attribute(section.get_attributes()[1])
        self.assertEqual(
            section.get_attributes(),
            [
                ["name1", "value1"],
                ["name3", "value3"],
            ]
        )

        section.del_attribute(["name3", "value3"])
        self.assertEqual(
            section.get_attributes(),
            [
                ["name1", "value1"],
            ]
        )

        section.del_attribute(["name3", "value3"])
        self.assertEqual(
            section.get_attributes(),
            [
                ["name1", "value1"],
            ]
        )

    def test_attribute_del_by_name(self):
        section = config_parser.Section("mySection")
        section.add_attribute("name1", "value1")
        section.add_attribute("name2", "value2")
        section.add_attribute("name3", "value3")
        section.add_attribute("name2", "value2")

        section.del_attributes_by_name("nameX")
        self.assertEqual(
            section.get_attributes(),
            [
                ["name1", "value1"],
                ["name2", "value2"],
                ["name3", "value3"],
                ["name2", "value2"],
            ]
        )

        section.del_attributes_by_name("name2", "value2")
        self.assertEqual(
            section.get_attributes(),
            [
                ["name1", "value1"],
                ["name3", "value3"],
            ]
        )

        section.add_attribute("name2", "value2")
        section.add_attribute("name2", "value2a")
        self.assertEqual(
            section.get_attributes(),
            [
                ["name1", "value1"],
                ["name3", "value3"],
                ["name2", "value2"],
                ["name2", "value2a"],
            ]
        )
        section.del_attributes_by_name("name2", "value2")
        self.assertEqual(
            section.get_attributes(),
            [
                ["name1", "value1"],
                ["name3", "value3"],
                ["name2", "value2a"],
            ]
        )

        section.add_attribute("name3", "value3a")
        self.assertEqual(
            section.get_attributes(),
            [
                ["name1", "value1"],
                ["name3", "value3"],
                ["name2", "value2a"],
                ["name3", "value3a"],
            ]
        )
        section.del_attributes_by_name("name3")
        self.assertEqual(
            section.get_attributes(),
            [
                ["name1", "value1"],
                ["name2", "value2a"],
            ]
        )

    def test_section_add(self):
        root = config_parser.Section("root")
        child1 = config_parser.Section("child1")
        child1a = config_parser.Section("child1a")
        child2 = config_parser.Section("child2")

        root.add_section(child1)
        child1.add_section(child1a)
        root.add_section(child2)
        self.assertEqual(root.parent, None)
        self.assertEqual(child1.parent.name, "root")
        self.assertEqual(child1a.parent.name, "child1")
        self.assertEqual(child2.parent.name, "root")
        ac(str(root), """\
child1 {
    child1a {
    }
}

child2 {
}
""")

        child2.add_section(child1a)
        self.assertEqual(child1a.parent.name, "child2")
        ac(str(root), """\
child1 {
}

child2 {
    child1a {
    }
}
""")

        self.assertRaises(
            config_parser.CircularParentshipException,
            child1a.add_section, child1a
        )
        self.assertRaises(
            config_parser.CircularParentshipException,
            child1a.add_section, child2
        )
        self.assertRaises(
            config_parser.CircularParentshipException,
            child1a.add_section, root
        )

    def test_section_get(self):
        root = config_parser.Section("")
        child1 = config_parser.Section("child1")
        child2 = config_parser.Section("child2")
        childa1 = config_parser.Section("childA")
        childa2 = config_parser.Section("childA")
        childa3 = config_parser.Section("childA")
        childa4 = config_parser.Section("childA")
        childb1 = config_parser.Section("childB")
        childb2 = config_parser.Section("childB")
        childa1.add_attribute("id", "1")
        childa2.add_attribute("id", "2")
        childa3.add_attribute("id", "3")
        childa4.add_attribute("id", "4")
        childb1.add_attribute("id", "5")
        childb2.add_attribute("id", "6")
        root.add_section(child1)
        root.add_section(child2)
        child1.add_section(childa1)
        child1.add_section(childa2)
        child1.add_section(childb1)
        child2.add_section(childa3)
        child2.add_section(childb2)
        child2.add_section(childa4)
        ac(str(root), """\
child1 {
    childA {
        id: 1
    }

    childA {
        id: 2
    }

    childB {
        id: 5
    }
}

child2 {
    childA {
        id: 3
    }

    childB {
        id: 6
    }

    childA {
        id: 4
    }
}
""")

        ac(
            "---\n".join([str(x) for x in root.get_sections()]),
            """\
child1 {
    childA {
        id: 1
    }

    childA {
        id: 2
    }

    childB {
        id: 5
    }
}
---
child2 {
    childA {
        id: 3
    }

    childB {
        id: 6
    }

    childA {
        id: 4
    }
}
""")

        ac(
            "---\n".join([str(x) for x in root.get_sections("child1")]),
            """\
child1 {
    childA {
        id: 1
    }

    childA {
        id: 2
    }

    childB {
        id: 5
    }
}
""")

        ac(
            "---\n".join([str(x) for x in child1.get_sections("childA")]),
            """\
childA {
    id: 1
}
---
childA {
    id: 2
}
""")

        ac(
            "---\n".join([str(x) for x in child1.get_sections("child2")]),
            ""
        )

    def test_section_del(self):
        root = config_parser.Section("")
        child1 = config_parser.Section("child1")
        child2 = config_parser.Section("child2")
        childa1 = config_parser.Section("childA")
        childa2 = config_parser.Section("childA")
        childa3 = config_parser.Section("childA")
        childa4 = config_parser.Section("childA")
        childb1 = config_parser.Section("childB")
        childb2 = config_parser.Section("childB")
        childa1.add_attribute("id", "1")
        childa2.add_attribute("id", "2")
        childa3.add_attribute("id", "3")
        childa4.add_attribute("id", "4")
        childb1.add_attribute("id", "5")
        childb2.add_attribute("id", "6")
        root.add_section(child1)
        root.add_section(child2)
        child1.add_section(childa1)
        child1.add_section(childa2)
        child1.add_section(childb1)
        child2.add_section(childa3)
        child2.add_section(childb2)
        child2.add_section(childa4)
        ac(str(root), """\
child1 {
    childA {
        id: 1
    }

    childA {
        id: 2
    }

    childB {
        id: 5
    }
}

child2 {
    childA {
        id: 3
    }

    childB {
        id: 6
    }

    childA {
        id: 4
    }
}
""")

        child2.del_section(childb2)
        self.assertEqual(childb2.parent, None)
        ac(str(root), """\
child1 {
    childA {
        id: 1
    }

    childA {
        id: 2
    }

    childB {
        id: 5
    }
}

child2 {
    childA {
        id: 3
    }

    childA {
        id: 4
    }
}
""")

        root.del_section(child2)
        self.assertEqual(child2.parent, None)
        ac(str(root), """\
child1 {
    childA {
        id: 1
    }

    childA {
        id: 2
    }

    childB {
        id: 5
    }
}
""")

        self.assertRaises(ValueError, root.del_section, child2)

        self.assertEqual(childa1.parent.name, "child1")
        self.assertRaises(ValueError, child2.del_section, childa1)
        self.assertEqual(childa1.parent.name, "child1")

        child1.del_section(childb1)
        self.assertEqual(childb1.parent, None)
        ac(str(root), """\
child1 {
    childA {
        id: 1
    }

    childA {
        id: 2
    }
}
""")

        child1.del_section(childa1)
        self.assertEqual(childa1.parent, None)
        child1.del_section(childa2)
        self.assertEqual(childa2.parent, None)
        ac(str(root), """\
child1 {
}
""")

        root.del_section(child1)
        self.assertEqual(child1.parent, None)
        ac(str(root), "")

    def test_get_root(self):
        root = config_parser.Section("root")
        child1 = config_parser.Section("child1")
        child1a = config_parser.Section("child1a")
        root.add_section(child1)
        child1.add_section(child1a)

        self.assertEqual(root.get_root().name, "root")
        self.assertEqual(child1.get_root().name, "root")
        self.assertEqual(child1a.get_root().name, "root")

    def test_str(self):
        root = config_parser.Section("root")
        ac(str(root), "")

        root.add_attribute("name1", "value1")
        ac(str(root), "name1: value1\n")

        root.add_attribute("name2", "value2")
        root.add_attribute("name2", "value2a")
        root.add_attribute("name3", "value3")
        ac(str(root), """\
name1: value1
name2: value2
name2: value2a
name3: value3
""")

        child1 = config_parser.Section("child1")
        root.add_section(child1)
        ac(str(root), """\
name1: value1
name2: value2
name2: value2a
name3: value3

child1 {
}
""")

        child1.add_attribute("name1.1", "value1.1")
        child1.add_attribute("name1.2", "value1.2")
        ac(str(root), """\
name1: value1
name2: value2
name2: value2a
name3: value3

child1 {
    name1.1: value1.1
    name1.2: value1.2
}
""")

        child2 = config_parser.Section("child2")
        child2.add_attribute("name2.1", "value2.1")
        root.add_section(child2)
        ac(str(root), """\
name1: value1
name2: value2
name2: value2a
name3: value3

child1 {
    name1.1: value1.1
    name1.2: value1.2
}

child2 {
    name2.1: value2.1
}
""")

        child2a = config_parser.Section("child2a")
        child2a.add_attribute("name2.a.1", "value2.a.1")
        child2.add_section(child2a)
        ac(str(root), """\
name1: value1
name2: value2
name2: value2a
name3: value3

child1 {
    name1.1: value1.1
    name1.2: value1.2
}

child2 {
    name2.1: value2.1

    child2a {
        name2.a.1: value2.a.1
    }
}
""")

        child3 = config_parser.Section("child3")
        root.add_section(child3)
        child3.add_section(config_parser.Section("child3a"))
        child3.add_section(config_parser.Section("child3b"))
        ac(str(root), """\
name1: value1
name2: value2
name2: value2a
name3: value3

child1 {
    name1.1: value1.1
    name1.2: value1.2
}

child2 {
    name2.1: value2.1

    child2a {
        name2.a.1: value2.a.1
    }
}

child3 {
    child3a {
    }

    child3b {
    }
}
""")


class ParserTest(unittest.TestCase):

    def test_empty(self):
        ac(str(config_parser.parse_string("")), "")

    def test_attributes(self):
        string = """\
name:value\
"""
        parsed = """\
name: value
"""
        ac(str(config_parser.parse_string(string)), parsed)

        string = """\
name:value
name:value
"""
        parsed = """\
name: value
name: value
"""
        ac(str(config_parser.parse_string(string)), parsed)

        string = """\
  name1:value1  
name2  :value2
name3:  value3
  name4  :  value4  
"""
        parsed = """\
name1: value1
name2: value2
name3: value3
name4: value4
"""
        ac(str(config_parser.parse_string(string)), parsed)

        string = """\
name:foo:value
"""
        parsed = """\
name: foo:value
"""
        root = config_parser.parse_string(string)
        self.assertEqual(root.get_attributes(), [["name", "foo:value"]])
        ac(str(root), parsed)

        string = """\
name :  
"""
        parsed = """\
name: 
"""
        root = config_parser.parse_string(string)
        self.assertEqual(root.get_attributes(), [["name", ""]])
        ac(str(root), parsed)

    def test_section(self):
        string = """\
section1 {
}\
"""
        parsed = """\
section1 {
}
"""
        ac(str(config_parser.parse_string(string)), parsed)

        string = """\
section1 {
    section1a   {
  }
  section1b        {       
     }    
}
"""
        parsed = """\
section1 {
    section1a {
    }

    section1b {
    }
}
"""
        ac(str(config_parser.parse_string(string)), parsed)

        string = """\
section1 {
    section1a junk1 { junk2
    junk3 } junk4
    section1b junk5{junk6
    junk7}junk8
}
section2 {
   section2a {
   }
   section2b {
   }
}
"""
        parsed = """\
section1 {
    section1a junk1 {
    }

    section1b junk5 {
    }
}

section2 {
    section2a {
    }

    section2b {
    }
}
"""
        ac(str(config_parser.parse_string(string)), parsed)

        string = """\
section1 {
    section1a {
    }

    section1b {
    }
}
}
"""
        self.assertRaises(
            config_parser.UnexpectedClosingBraceException,
            config_parser.parse_string, string
        )

        string = """\
section1 {
    section1a {

    section1b {
    }
}
"""
        self.assertRaises(
            config_parser.MissingClosingBraceException,
            config_parser.parse_string, string
        )

        string = """\
section1 {
"""
        self.assertRaises(
            config_parser.MissingClosingBraceException,
            config_parser.parse_string, string
        )

        string = """\
}
"""
        self.assertRaises(
            config_parser.UnexpectedClosingBraceException,
            config_parser.parse_string, string
        )


    def test_comment(self):
        string= """\
# junk1
name1: value1
  #junk2
name2: value2#junk3
name3: value3 #junk4
name4 # junk5: value4
#junk6 name5: value5
#junk7
"""
        parsed = """\
name1: value1
name2: value2#junk3
name3: value3 #junk4
name4 # junk5: value4
"""
        ac(str(config_parser.parse_string(string)), parsed)

        string= """\
# junk1
section1 { # junk2
}
section2 # junk2 {
}
section3 {
} #junk3
"""
        parsed = """\
section1 {
}

section2 # junk2 {
}

section3 {
}
"""
        ac(str(config_parser.parse_string(string)), parsed)

        string = """\
section {
#}
"""
        self.assertRaises(
            config_parser.MissingClosingBraceException,
            config_parser.parse_string, string
        )

        string = """\
#section {
}
"""
        self.assertRaises(
            config_parser.UnexpectedClosingBraceException,
            config_parser.parse_string, string
        )

    def test_full(self):
        string = """\
# Please read the corosync.conf.5 manual page
totem {
	version: 2

	# crypto_cipher and crypto_hash: Used for mutual node authentication.
	# If you choose to enable this, then do remember to create a shared
	# secret with "corosync-keygen".
	# enabling crypto_cipher, requires also enabling of crypto_hash.
	crypto_cipher: none
	crypto_hash: none

	# interface: define at least one interface to communicate
	# over. If you define more than one interface stanza, you must
	# also set rrp_mode.
	interface {
                # Rings must be consecutively numbered, starting at 0.
		ringnumber: 0
		# This is normally the *network* address of the
		# interface to bind to. This ensures that you can use
		# identical instances of this configuration file
		# across all your cluster nodes, without having to
		# modify this option.
		bindnetaddr: 192.168.1.0
		# However, if you have multiple physical network
		# interfaces configured for the same subnet, then the
		# network address alone is not sufficient to identify
		# the interface Corosync should bind to. In that case,
		# configure the *host* address of the interface
		# instead:
		# bindnetaddr: 192.168.1.1
		# When selecting a multicast address, consider RFC
		# 2365 (which, among other things, specifies that
		# 239.255.x.x addresses are left to the discretion of
		# the network administrator). Do not reuse multicast
		# addresses across multiple Corosync clusters sharing
		# the same network.
		mcastaddr: 239.255.1.1
		# Corosync uses the port you specify here for UDP
		# messaging, and also the immediately preceding
		# port. Thus if you set this to 5405, Corosync sends
		# messages over UDP ports 5405 and 5404.
		mcastport: 5405
		# Time-to-live for cluster communication packets. The
		# number of hops (routers) that this ring will allow
		# itself to pass. Note that multicast routing must be
		# specifically enabled on most network routers.
		ttl: 1
	}
}

logging {
	# Log the source file and line where messages are being
	# generated. When in doubt, leave off. Potentially useful for
	# debugging.
	fileline: off
	# Log to standard error. When in doubt, set to no. Useful when
	# running in the foreground (when invoking "corosync -f")
	to_stderr: no
	# Log to a log file. When set to "no", the "logfile" option
	# must not be set.
	to_logfile: yes
	logfile: /var/log/cluster/corosync.log
	# Log to the system log daemon. When in doubt, set to yes.
	to_syslog: yes
	# Log debug messages (very verbose). When in doubt, leave off.
	debug: off
	# Log messages with time stamps. When in doubt, set to on
	# (unless you are only logging to syslog, where double
	# timestamps can be annoying).
	timestamp: on
	logger_subsys {
		subsys: QUORUM
		debug: off
	}
}

quorum {
	# Enable and configure quorum subsystem (default: off)
	# see also corosync.conf.5 and votequorum.5
	#provider: corosync_votequorum
}
"""
        parsed = """\
totem {
    version: 2
    crypto_cipher: none
    crypto_hash: none

    interface {
        ringnumber: 0
        bindnetaddr: 192.168.1.0
        mcastaddr: 239.255.1.1
        mcastport: 5405
        ttl: 1
    }
}

logging {
    fileline: off
    to_stderr: no
    to_logfile: yes
    logfile: /var/log/cluster/corosync.log
    to_syslog: yes
    debug: off
    timestamp: on

    logger_subsys {
        subsys: QUORUM
        debug: off
    }
}

quorum {
}
"""
        ac(str(config_parser.parse_string(string)), parsed)

        string = """\
# Please read the corosync.conf.5 manual page
totem {
	version: 2

	crypto_cipher: none
	crypto_hash: none

	interface {
		ringnumber: 0
		bindnetaddr: 10.16.35.0
		mcastport: 5405
		ttl: 1
	}
	transport: udpu
}

logging {
	fileline: off
	to_logfile: yes
	to_syslog: yes
	logfile: /var/log/cluster/corosync.log
	debug: off
	timestamp: on
	logger_subsys {
		subsys: QUORUM
		debug: off
	}
}

nodelist {
	node {
		ring0_addr: 10.16.35.101
		nodeid: 1
	}

	node {
		ring0_addr: 10.16.35.102
		nodeid: 2
	}

	node {
		ring0_addr: 10.16.35.103
	}

	node {
		ring0_addr: 10.16.35.104
	}

	node {
		ring0_addr: 10.16.35.105
	}
}

quorum {
	# Enable and configure quorum subsystem (default: off)
	# see also corosync.conf.5 and votequorum.5
	#provider: corosync_votequorum
}
"""
        parsed = """\
totem {
    version: 2
    crypto_cipher: none
    crypto_hash: none
    transport: udpu

    interface {
        ringnumber: 0
        bindnetaddr: 10.16.35.0
        mcastport: 5405
        ttl: 1
    }
}

logging {
    fileline: off
    to_logfile: yes
    to_syslog: yes
    logfile: /var/log/cluster/corosync.log
    debug: off
    timestamp: on

    logger_subsys {
        subsys: QUORUM
        debug: off
    }
}

nodelist {
    node {
        ring0_addr: 10.16.35.101
        nodeid: 1
    }

    node {
        ring0_addr: 10.16.35.102
        nodeid: 2
    }

    node {
        ring0_addr: 10.16.35.103
    }

    node {
        ring0_addr: 10.16.35.104
    }

    node {
        ring0_addr: 10.16.35.105
    }
}

quorum {
}
"""
        ac(str(config_parser.parse_string(string)), parsed)



from __future__ import (
    absolute_import,
    division,
    print_function,
)
from pcs.cli.constraint.console_report import prepare_options


def constraint_plain(constraint_info, with_id=False):
    """
    dict constraint_info  see constraint in pcs/lib/exchange_formats.md
    bool with_id have to show id with options_dict
    """
    options = constraint_info["options"]
    role = options.get("rsc-role", "")
    role_prefix = "{0} ".format(role) if role else ""

    return role_prefix + " ".join([options.get("rsc", "")] + prepare_options(
        dict(
            (name, value) for name, value in options.items()
            if name not in ["rsc-role", "rsc"]
        ),
        with_id
    ))



from __future__ import (
    absolute_import,
    division,
    print_function,
)

from pcs.test.tools.pcs_unittest import TestCase

from pcs.test.tools.assertions import (
    assert_raise_library_error,
    assert_xml_equal,
)
from pcs.test.tools.misc import outdent

from lxml import etree

from pcs.common import report_codes
from pcs.lib.errors import ReportItemSeverity as severity

from pcs.lib import cluster_conf_facade as lib

class FromStringTest(TestCase):
    def test_success(self):
        facade = lib.ClusterConfFacade.from_string("<cluster/>")
        self.assertTrue(isinstance(facade, lib.ClusterConfFacade))
        assert_xml_equal("<cluster/>", etree.tostring(facade._config).decode())

    def test_syntax_error(self):
        assert_raise_library_error(
            lambda: lib.ClusterConfFacade.from_string("<cluster>"),
            (
                severity.ERROR,
                report_codes.CLUSTER_CONF_LOAD_ERROR_INVALID_FORMAT,
                {}
            )
        )

    def test_invalid_document_error(self):
        assert_raise_library_error(
            lambda: lib.ClusterConfFacade.from_string("string"),
            (
                severity.ERROR,
                report_codes.CLUSTER_CONF_LOAD_ERROR_INVALID_FORMAT,
                {}
            )
        )


class GetClusterNameTest(TestCase):
    def test_success(self):
        cfg = etree.XML('<cluster name="cluster-name"/>')
        self.assertEqual(
            "cluster-name",
            lib.ClusterConfFacade(cfg).get_cluster_name()
        )

    def test_no_name(self):
        cfg = etree.XML('<cluster/>')
        self.assertEqual("", lib.ClusterConfFacade(cfg).get_cluster_name())

    def test_not_cluster_element(self):
        cfg = etree.XML('<not_cluster/>')
        self.assertEqual("", lib.ClusterConfFacade(cfg).get_cluster_name())


class GetNodesTest(TestCase):
    def assert_equal_nodelist(self, expected_nodes, real_nodelist):
        real_nodes = [
            {"ring0": n.ring0, "ring1": n.ring1, "name": n.name, "id": n.id}
            for n in real_nodelist
        ]
        self.assertEqual(expected_nodes, real_nodes)

    def test_success(self):
        config = outdent("""
            <cluster>
                <clusternodes>
                    <clusternode name="node1" nodeid="1">
                        <altname name="node1-altname"/>
                    </clusternode>
                    <clusternode name="node2" nodeid="2"/>
                    <clusternode name="node3" nodeid="3"/>
                </clusternodes>
            </cluster>
        """)
        self.assert_equal_nodelist(
            [
                {
                    "ring0": "node1",
                    "ring1": "node1-altname",
                    "name": None,
                    "id": "1",
                },
                {
                    "ring0": "node2",
                    "ring1": None,
                    "name": None,
                    "id": "2",
                },
                {
                    "ring0": "node3",
                    "ring1": None,
                    "name": None,
                    "id": "3",
                }
            ],
            lib.ClusterConfFacade(etree.XML(config)).get_nodes()
        )

    def test_no_nodes(self):
        config = "<cluster/>"
        self.assert_equal_nodelist(
            [], lib.ClusterConfFacade(etree.XML(config)).get_nodes()
        )

    def test_missing_info(self):
        config = outdent("""
            <cluster>
                <clusternodes>
                    <clusternode nodeid="1"/>
                    <clusternode name="node2">
                        <altname/>
                    </clusternode>
                    <clusternode/>
                </clusternodes>
            </cluster>
        """)
        self.assert_equal_nodelist(
            [
                {
                    "ring0": None,
                    "ring1": None,
                    "name": None,
                    "id": "1",
                },
                {
                    "ring0": "node2",
                    "ring1": None,
                    "name": None,
                    "id": None,
                },
                {
                    "ring0": None,
                    "ring1": None,
                    "name": None,
                    "id": None,
                }
            ],
            lib.ClusterConfFacade(etree.XML(config)).get_nodes()
        )



from __future__ import (
    absolute_import,
    division,
    print_function,
)

import json

try:
    # python 2
    from urlparse import parse_qs
except ImportError:
    # python 3
    from urllib.parse import parse_qs

from pcs.common import pcs_pycurl as pycurl
from pcs.common.node_communicator import(
    RequestTarget,
    RequestData,
    Request,
    Response,
)
from pcs.test.tools.custom_mock import MockCurlSimple

CALL_TYPE_HTTP_ADD_REQUESTS = "CALL_TYPE_HTTP_ADD_REQUESTS"
CALL_TYPE_HTTP_START_LOOP = "CALL_TYPE_HTTP_START_LOOP"

def log_request(request):
    label_data = [
            ("action", request.action),
            ("label", request.target.label),
            ("data", parse_qs(request.data)),
        ]

    if request.target.address_list != [request.target.label]:
        label_data.append(("address_list", request.target.address_list))

    if request.target.port != "2224":
        label_data.append(("port", request.target.port))


    return "  ".join([
        "{0}:'{1}'".format(key, value) for key, value in label_data
    ])

def log_response(response, indent=0):
    label_data = [
        ("action", response.request.action),
        ("label", response.request.target.label),
    ]

    if response.request.target.address_list != [response.request.target.label]:
        label_data.append(("address_list", response.request.target.address_list))

    label_data.append(("was_connected", response.was_connected))

    if response.request.target.port != "2224":
        label_data.append(("port", response.request.target.port))

    if response.was_connected:
        label_data.append(("respose_code", response.response_code))
    else:
        label_data.extend([
            ("errno", response.errno),
            ("error_msg", response.error_msg),
        ])

    label_data.append(("data", parse_qs(response.request.data)))

    return "{0}{1}".format(
        "  "*indent,
        "  ".join([
            "{0}:'{1}'".format(key, value) for key, value in label_data
        ]),
    )

def different_request_lists(expected_request_list, request_list):
    return AssertionError(
        (
            "Method add_request of NodeCommunicator expected"
            " request_list:\n  * {0}\nbut got: \n  * {1}"
        )
        .format(
            "\n  * ".join(log_request(r) for r in expected_request_list),
            "\n  * ".join(log_request(r) for r in request_list),
        )
    )

def bad_request_list_content(errors):
    return AssertionError(
        "Method add_request of NodeCommunicator get different requests"
        " than expected (key: (expected, real)): \n  {0}".format(
            "\n  ".join([
                "{0}:\n    {1}".format(
                    index,
                    "\n    ".join([
                        "{0}:\n      {1}\n      {2}"
                        .format(key, pair[0], pair[1])
                        for key, pair in value.items()
                    ])
                )
                for index, value in errors.items()
            ]),
        )
    )

def _communication_to_response(
    label, address_list, action, param_list, port, token, response_code,
    output, debug_output, was_connected, errno, error_msg
):
    return Response(
        MockCurlSimple(
            info={pycurl.RESPONSE_CODE: response_code},
            output=output,
            debug_output=debug_output,
            request=Request(
                RequestTarget(label, address_list, port, token),
                RequestData(action, param_list),
            )
        ),
        was_connected=was_connected,
        errno=errno,
        error_msg=error_msg,
    )

def create_communication(
    communication_list, action="", param_list=None, port=None, token=None,
    response_code=200, output="", debug_output="", was_connected=True,
    errno=0, error_msg_template=None
):
    """
    list of dict communication_list -- is setting for one request - response
        it accepts keys:
            label -- required, see RequestTarget
            action -- pcsd url, see RequestData
            param_list -- list of pairs, see RequestData
            port -- see RequestTarget
            token=None -- see RequestTarget
            response_code -- http response code
            output -- http response output
            debug_output -- pycurl debug output
            was_connected -- see Response
            errno -- see Response
            error_msg -- see Response
        if some key is not present, it is put here from common values - rest
        args of this fuction(except name, communication_list,
        error_msg_template)
    string error_msg_template -- template, the keys for format function will
        be taken from appropriate item of communication_list
    string action -- pcsd url, see RequestData
    list of pairs (tuple) param_list -- see RequestData
    string port -- see RequestTarget
    string token=None -- see RequestTarget
    string response_code -- http response code
    string output -- http response output
    string debug_output -- pycurl debug output
    bool was_connected -- see Response
    int errno -- see Response
    string error_msg -- see Response
    """
    response_list = []

    common = dict(
        action=action,
        param_list=param_list if param_list else (),
        port=port,
        token=token,
        response_code=response_code,
        output=output,
        debug_output=debug_output,
        was_connected=was_connected,
        errno=errno,
    )
    for communication in communication_list:
        if "address_list" not in communication:
            communication["address_list"] = [communication["label"]]

        full = common.copy()
        full.update(communication)

        if "error_msg" not in full:
            full["error_msg"] = (
                "" if not error_msg_template
                else error_msg_template.format(**full)
            )
        response_list.append(
            _communication_to_response(**full)
        )

    request_list = [response.request for response in response_list]

    return request_list, response_list

def place_multinode_call(
    calls, name, node_labels=None, communication_list=None, **kwargs
):
    """
    Shortcut for adding a call sending the same request to one or more nodes

    CallListBuilder calls -- list of expected calls
    string name -- the key of this call
    list node_labels -- create success responses from these nodes
    list communication_list -- use these custom responses
    **kwargs -- see __module__.create_communication
    """
    if (
        (not node_labels and not communication_list)
        or
        (node_labels and communication_list)
    ):
        raise AssertionError(
            "Exactly one of 'node_labels', 'communication_list' "
            "must be specified"
        )
    communication_list = communication_list or [
        {"label": label} for label in node_labels
    ]
    place_communication(calls, name, communication_list, **kwargs)


def place_requests(calls, name, request_list):
    calls.place(name, AddRequestCall(request_list))


def place_responses(calls, name, response_list):
    calls.place(name, StartLoopCall(response_list))


def place_communication(calls, name, communication_list, **kwargs):
    if isinstance(communication_list[0], dict):
        communication_list = [communication_list]

    request_list = []
    response_list = []
    for com_list in communication_list:
        req_list, res_list = create_communication(com_list, **kwargs)
        request_list.append(req_list)
        response_list.extend(res_list)

    place_requests(calls, "{0}_requests".format(name), request_list[0])
    place_responses(calls, "{0}_responses".format(name), response_list)
    for i, req_list in enumerate(request_list[1:], start=1):
        place_requests(calls, "{0}_requests_{1}".format(name, i), req_list)


class AddRequestCall(object):
    type = CALL_TYPE_HTTP_ADD_REQUESTS

    def __init__(self, request_list):
        self.request_list = request_list

    def format(self):
        return "Requests:\n    * {0}".format(
            "\n    * ".join([
                log_request(request) for request in self.request_list
            ])
        )

    def __repr__(self):
        return str("<HttpAddRequest '{0}'>").format(self.request_list)

class StartLoopCall(object):
    type = CALL_TYPE_HTTP_START_LOOP

    def format(self):
        return "Responses:\n    * {0}".format(
            "\n    * ".join([
                log_response(response) for response in self.response_list
            ])
        )

    def __init__(self, response_list):
        self.response_list = response_list

    def __repr__(self):
        return str("<HttpStartLoop '{0}'>").format(self.response_list)

def _compare_request_data(expected, real):
    if expected == real:
        return True

    # If data is in json format it is not possible to compare it as string.
    # Because python 3 does not keep key order of dict. So if is response
    # builded by json.dumps(some_dict) the result string can vary.

    # Let's try known use: [('data_json', 'some_json_here')]
    # It means only one pair "data_json" + json string: everything else is False

    if len(expected) != 1:
        return False

    if len(real) != 1:
        return False

    if expected[0][0] != real[0][0] or expected[0][0] != "data_json":
        return False

    try:
        expected_data = json.loads(expected[0][1])
        real_data = json.loads(real[0][1])
        return expected_data == real_data
    except ValueError:
        return False


class NodeCommunicator(object):
    def __init__(self, call_queue=None):
        self.__call_queue = call_queue

    def add_requests(self, request_list):
        _, add_request_call = self.__call_queue.take(
            CALL_TYPE_HTTP_ADD_REQUESTS,
            request_list,
        )

        expected_request_list = add_request_call.request_list

        if len(expected_request_list) != len(request_list):
            raise different_request_lists(expected_request_list, request_list)

        errors = {}
        for i, real_request in enumerate(request_list):
            expected_request = add_request_call.request_list[i]

            diff = {}
            if expected_request.action != real_request.action:
                diff["action"] = (
                    expected_request.action,
                    real_request.action
                )

            if expected_request.target.label != real_request.target.label:
                diff["target.label"] = (
                    expected_request.target.label,
                    real_request.target.label
                )

            if expected_request.target.token != real_request.target.token:
                diff["target.token"] = (
                    expected_request.target.token,
                    real_request.target.token
                )

            if expected_request.target.port != real_request.target.port:
                diff["target.port"] = (
                    expected_request.target.port,
                    real_request.target.port
                )

            if not _compare_request_data(
                expected_request._data.structured_data,
                real_request._data.structured_data
            ):
                diff["data"] = (
                    expected_request._data.structured_data,
                    real_request._data.structured_data,
                )

            if diff:
                errors[i] = diff

        if errors:
            raise self.__call_queue.error_with_context(
                bad_request_list_content(errors)
            )


    def start_loop(self):
        _, call = self.__call_queue.take(CALL_TYPE_HTTP_START_LOOP)
        return call.response_list



from __future__ import (
    absolute_import,
    division,
    print_function,
)

from collections import Iterable
from functools import partial
import sys

from pcs.common import report_codes as codes
from pcs.common.fencing_topology import TARGET_TYPE_ATTRIBUTE
from pcs.common.tools import is_string

INSTANCE_SUFFIX = "@{0}"
NODE_PREFIX = "{0}: "

_type_translation = {
    "acl_group": "ACL group",
    "acl_permission": "ACL permission",
    "acl_role": "ACL role",
    "acl_target": "ACL user",
    "primitive": "resource",
}
_type_articles = {
    "ACL group": "an",
    "ACL user": "an",
    "ACL role": "an",
    "ACL permission": "an",
}
_file_role_translation = {
    "BOOTH_CONFIG": "Booth configuration",
    "BOOTH_KEY": "Booth key",
    "COROSYNC_AUTHKEY": "Corosync authkey",
    "PACEMAKER_AUTHKEY": "Pacemaker authkey",
    "PCSD_SSL_CERT": "pcsd SSL certificate",
    "PCSD_SSL_KEY": "pcsd SSL key",
    "PCS_SETTINGS_CONF": "pcs configuration",
}

def warn(message):
    sys.stdout.write(format_message(message, "Warning: "))

def format_message(message, prefix):
    return "{0}{1}\n".format(prefix, message)

def error(message):
    sys.stderr.write(format_message(message, "Error: "))
    return SystemExit(1)

def indent(line_list, indent_step=2):
    """
    return line list where each line of input is prefixed by N spaces
    list of string line_list are original lines
    int indent_step is count of spaces for line prefix
    """
    return [
        "{0}{1}".format(" "*indent_step, line) if line else line
        for line in line_list
    ]

def format_optional(value, template, empty_case=""):
    return empty_case if not value else template.format(value)

def format_fencing_level_target(target_type, target_value):
    if target_type == TARGET_TYPE_ATTRIBUTE:
        return "{0}={1}".format(target_value[0], target_value[1])
    return target_value

def format_file_role(role):
    return _file_role_translation.get(role, role)

def service_operation_started(operation, info):
    return "{operation} {service}{instance_suffix}...".format(
        operation=operation,
        instance_suffix=format_optional(info["instance"], INSTANCE_SUFFIX),
        **info
    )

def service_operation_error(operation, info):
    return (
        "{node_prefix}Unable to {operation} {service}{instance_suffix}:"
        " {reason}"
    ).format(
        operation=operation,
        instance_suffix=format_optional(info["instance"], INSTANCE_SUFFIX),
        node_prefix=format_optional(info["node"], NODE_PREFIX),
        **info
    )

def service_operation_success(operation, info):
    return "{node_prefix}{service}{instance_suffix} {operation}".format(
        operation=operation,
        instance_suffix=format_optional(info["instance"], INSTANCE_SUFFIX),
        node_prefix=format_optional(info["node"], NODE_PREFIX),
        **info
    )

def service_operation_skipped(operation, info):
    return (
        "{node_prefix}not {operation} {service}{instance_suffix}: {reason}"
    ).format(
        operation=operation,
        instance_suffix=format_optional(info["instance"], INSTANCE_SUFFIX),
        node_prefix=format_optional(info["node"], NODE_PREFIX),
        **info
    )

def typelist_to_string(type_list, article=False):
    if not type_list:
        return ""
    new_list = sorted([
        # get a translation or make a type_name a string
        _type_translation.get(type_name, "{0}".format(type_name))
        for type_name in type_list
    ])
    types = "/".join(new_list)
    if not article:
        return types
    return "{article} {types}".format(
        article=_type_articles.get(new_list[0], "a"),
        types=types
    )

def id_belongs_to_unexpected_type(info):
    return "'{id}' is not {expected_type}".format(
        id=info["id"],
        expected_type=typelist_to_string(info["expected_types"], article=True)
    )

def id_not_found(info):
    desc = format_optional(typelist_to_string(info["expected_types"]), "{0} ")
    if not info["context_type"] or not info["context_id"]:
        return "{desc}'{id}' does not exist".format(desc=desc, id=info["id"])

    return (
        "there is no {desc}'{id}' in the {context_type} '{context_id}'".format(
            desc=desc,
            id=info["id"],
            context_type=info["context_type"],
            context_id=info["context_id"],
        )
    )

def resource_running_on_nodes(info):
    role_label_map = {
        "Started": "running",
    }
    state_info = {}
    for state, node_list in info["roles_with_nodes"].items():
        state_info.setdefault(
            role_label_map.get(state, state.lower()),
            []
        ).extend(node_list)

    return "resource '{resource_id}' is {detail_list}".format(
        resource_id=info["resource_id"],
        detail_list="; ".join(sorted([
            "{run_type} on node{s} {node_list}".format(
                run_type=run_type,
                s="s" if len(node_list) > 1 else "",
                node_list=joined_list(node_list)
            )
            for run_type, node_list in state_info.items()
        ]))
    )

def invalid_options(info):
    template = "invalid {desc}option{plural_options} {option_names_list},"
    if not info["allowed"] and not info["allowed_patterns"]:
        template += " there are no options allowed"
    elif not info["allowed_patterns"]:
        template += " allowed option{plural_allowed} {allowed_values}"
    elif not info["allowed"]:
        template += (
            " allowed are options matching patterns: {allowed_patterns_values}"
        )
    else:
        template += (
            " allowed option{plural_allowed} {allowed_values}"
            " and"
            " options matching patterns: {allowed_patterns_values}"
        )
    return template.format(
        desc=format_optional(info["option_type"], "{0} "),
        allowed_values=", ".join(sorted(info["allowed"])),
        allowed_patterns_values=", ".join(sorted(info["allowed_patterns"])),
        option_names_list=joined_list(info["option_names"]),
        plural_options=("s:" if len(info["option_names"]) > 1 else ""),
        plural_allowed=("s are:" if len(info["allowed"]) > 1 else " is"),
        **info
    )

def build_node_description(node_types):
    if not node_types:
        return  "Node"

    label = "{0} node".format

    if is_string(node_types):
        return label(node_types)

    if len(node_types) == 1:
        return label(node_types[0])

    return "nor " + " or ".join([label(ntype) for ntype in node_types])

def joined_list(item_list, optional_transformations=None):
    if not optional_transformations:
        optional_transformations={}

    return ", ".join(sorted([
        "'{0}'".format(optional_transformations.get(item, item))
        for item in item_list
    ]))

#Each value (a callable taking report_item.info) returns a message.
#Force text will be appended if necessary.
#If it is necessary to put the force text inside the string then the callable
#must take the force_text parameter.
CODE_TO_MESSAGE_BUILDER_MAP = {

    codes.COMMON_ERROR: lambda info: info["text"],

    codes.COMMON_INFO: lambda info: info["text"],

    codes.EMPTY_RESOURCE_SET_LIST: "Resource set list is empty",

    codes.REQUIRED_OPTION_IS_MISSING: lambda info:
        "required {desc}option{s} {option_names_list} {are} missing"
        .format(
            desc=format_optional(info["option_type"], "{0} "),
            option_names_list=joined_list(info["option_names"]),
            s=("s" if len(info["option_names"]) > 1 else ""),
            are=(
                "are" if len(info["option_names"]) > 1
                else "is"
            )
        )
    ,

    codes.PREREQUISITE_OPTION_IS_MISSING: lambda info:
        (
            "If {opt_desc}option '{option_name}' is specified, "
            "{pre_desc}option '{prerequisite_name}' must be specified as well"
        ).format(
            opt_desc=format_optional(info.get("option_type"), "{0} "),
            pre_desc=format_optional(info.get("prerequisite_type"), "{0} "),
            **info
        )
    ,

    codes.REQUIRED_OPTION_OF_ALTERNATIVES_IS_MISSING: lambda info:
        "{desc}option {option_names_list} has to be specified"
        .format(
            desc=format_optional(info.get("option_type"), "{0} "),
            option_names_list=" or ".join(sorted([
                "'{0}'".format(name)
                for name in info["option_names"]
            ])),
        )
    ,

    codes.INVALID_OPTIONS: invalid_options,

    codes.INVALID_OPTION_VALUE: lambda info:
        #value on key "allowed_values" is overloaded:
        # * it can be a list - then it express possible option values
        # * it can be a string - then it is verbal description of value
        "'{option_value}' is not a valid {option_name} value, use {hint}"
        .format(
            hint=(
                ", ".join(sorted(info["allowed_values"])) if (
                    isinstance(info["allowed_values"], Iterable)
                    and
                    not is_string(info["allowed_values"])
                ) else info["allowed_values"]
            ),
            **info
        )
    ,

    codes.INVALID_OPTION_TYPE: lambda info:
        #value on key "allowed_types" is overloaded:
        # * it can be a list - then it express possible option types
        # * it can be a string - then it is verbal description of the type
        "specified {option_name} is not valid, use {hint}"
        .format(
            hint=(
                ", ".join(sorted(info["allowed_types"])) if (
                    isinstance(info["allowed_types"], Iterable)
                    and
                    not is_string(info["allowed_types"])
                ) else info["allowed_types"]
            ),
            **info
        )
    ,

    codes.INVALID_USERDEFINED_OPTIONS: lambda info:
        (
            "invalid {desc}option{plural_options} {option_names_list}, "
            "{allowed_description}"
        ).format(
            desc=format_optional(info["option_type"], "{0} "),
            option_names_list=joined_list(info["option_names"]),
            plural_options=("s:" if len(info["option_names"]) > 1 else ""),
            **info
        )
    ,

    codes.DEPRECATED_OPTION: lambda info:
        (
            "{desc}option '{option_name}' is deprecated and should not be "
            "used, use {hint} instead"
        ).format(
            desc=format_optional(info["option_type"], "{0} "),
            hint=(
                ", ".join(sorted(info["replaced_by"])) if (
                    isinstance(info["replaced_by"], Iterable)
                    and
                    not is_string(info["replaced_by"])
                ) else info["replaced_by"]
            ),
            **info
        )
    ,

    codes.MUTUALLY_EXCLUSIVE_OPTIONS: lambda info:
        # "{desc}options {option_names} are muttually exclusive".format(
        "Only one of {desc}options {option_names} can be used".format(
            desc=format_optional(info["option_type"], "{0} "),
            option_names=(
                joined_list(sorted(info["option_names"])[:-1])
                +
                " and '{0}'".format(sorted(info["option_names"])[-1])
            )
        )
    ,

    codes.EMPTY_ID: lambda info:
        "{id_description} cannot be empty"
        .format(**info)
    ,

    codes.INVALID_CIB_CONTENT: lambda info:
        "invalid cib: \n{0}"
        .format(info["report"])
    ,

    codes.INVALID_ID: lambda info:
        (
            "invalid {id_description} '{id}', '{invalid_character}' "
            "is not a valid {desc}character for a {id_description}"
        ).format(
            desc="first " if info["is_first_char"] else "",
            **info
        )
    ,

    codes.INVALID_TIMEOUT_VALUE: lambda info:
        "'{timeout}' is not a valid number of seconds to wait"
        .format(**info)
    ,


    codes.INVALID_SCORE: lambda info:
        "invalid score '{score}', use integer or INFINITY or -INFINITY"
        .format(**info)
    ,


    codes.MULTIPLE_SCORE_OPTIONS: "you cannot specify multiple score options",


    codes.RUN_EXTERNAL_PROCESS_STARTED: lambda info:
        "Running: {command}\nEnvironment:{env_part}\n{stdin_part}".format(
            stdin_part=format_optional(
                info["stdin"],
                "--Debug Input Start--\n{0}\n--Debug Input End--\n"
            ),
            env_part=(
                "" if not info["environment"] else "\n" + "\n".join([
                    "  {0}={1}".format(key, val)
                    for key, val in sorted(info["environment"].items())
                ])
            ),
            **info
        )
    ,

    codes.RUN_EXTERNAL_PROCESS_FINISHED: lambda info:
        (
            "Finished running: {command}\n"
            "Return value: {return_value}\n"
            "--Debug Stdout Start--\n"
            "{stdout}\n"
            "--Debug Stdout End--\n"
            "--Debug Stderr Start--\n"
            "{stderr}\n"
            "--Debug Stderr End--\n"
        ).format(**info)
    ,

    codes.RUN_EXTERNAL_PROCESS_ERROR: lambda info:
        "unable to run command {command}: {reason}"
        .format(**info)
    ,

    codes.NODE_COMMUNICATION_DEBUG_INFO: lambda info:
        (
            "Communication debug info for calling: {target}\n"
            "--Debug Communication Info Start--\n"
            "{data}\n"
            "--Debug Communication Info End--\n"
        ).format(**info)
    ,

    codes.NODE_COMMUNICATION_STARTED: lambda info:
        "Sending HTTP Request to: {target}\n{data_part}".format(
            data_part=format_optional(
                info["data"],
                "--Debug Input Start--\n{0}\n--Debug Input End--\n"
            ),
            **info
        )
    ,

    codes.NODE_COMMUNICATION_FINISHED: lambda info:
        (
            "Finished calling: {target}\n"
            "Response Code: {response_code}\n"
            "--Debug Response Start--\n"
            "{response_data}\n"
            "--Debug Response End--\n"
        ).format(**info)
    ,

    codes.NODE_COMMUNICATION_NOT_CONNECTED: lambda info:
        "Unable to connect to {node} ({reason})"
        .format(**info)
    ,

    codes.NODE_COMMUNICATION_ERROR_NOT_AUTHORIZED: lambda info:
        (
            "Unable to authenticate to {node} ({reason}),"
            " try running 'pcs cluster auth'"
        )
        .format(**info)
    ,

    codes.NODE_COMMUNICATION_ERROR_PERMISSION_DENIED: lambda info:
        "{node}: Permission denied ({reason})"
        .format(**info)
    ,

    codes.NODE_COMMUNICATION_ERROR_UNSUPPORTED_COMMAND: lambda info:
        "{node}: Unsupported command ({reason}), try upgrading pcsd"
        .format(**info)
    ,

    codes.NODE_COMMUNICATION_COMMAND_UNSUCCESSFUL: lambda info:
        "{node}: {reason}"
        .format(**info)
    ,

    codes.NODE_COMMUNICATION_ERROR: lambda info:
        "Error connecting to {node} ({reason})"
        .format(**info)
    ,

    codes.NODE_COMMUNICATION_ERROR_UNABLE_TO_CONNECT: lambda info:
        "Unable to connect to {node} ({reason})"
        .format(**info)
    ,

    codes.NODE_COMMUNICATION_ERROR_TIMED_OUT: lambda info:
        (
            "{node}: Connection timeout, try setting higher timeout in "
            "--request-timeout option ({reason})"
        ).format(**info)
    ,

    codes.NODE_COMMUNICATION_PROXY_IS_SET:
        "Proxy is set in environment variables, try disabling it"
    ,

    codes.CANNOT_ADD_NODE_IS_IN_CLUSTER: lambda info:
        "cannot add the node '{node}' because it is in a cluster"
        .format(**info)
    ,

    codes.CANNOT_ADD_NODE_IS_RUNNING_SERVICE: lambda info:
        (
            "cannot add the node '{node}' because it is running service"
            " '{service}'{guess}"
        ).format(
            guess=(
                "" if info["service"] not in ["pacemaker", "pacemaker_remote"]
                else " (is not the node already in a cluster?)"
            ),
            **info
        )
    ,

    codes.DEFAULTS_CAN_BE_OVERRIDEN:
        "Defaults do not apply to resources which override them with their "
        "own defined values"
    ,

    codes.COROSYNC_CONFIG_DISTRIBUTION_STARTED:
        "Sending updated corosync.conf to nodes..."
    ,

    codes.COROSYNC_CONFIG_ACCEPTED_BY_NODE: lambda info:
        "{node}: Succeeded"
        .format(**info)
    ,

    codes.COROSYNC_CONFIG_DISTRIBUTION_NODE_ERROR: lambda info:
        "{node}: Unable to set corosync config"
        .format(**info)
    ,

    codes.COROSYNC_NOT_RUNNING_CHECK_STARTED:
        "Checking corosync is not running on nodes..."
    ,

    codes.COROSYNC_NOT_RUNNING_CHECK_NODE_ERROR: lambda info:
        "{node}: Unable to check if corosync is not running"
        .format(**info)
    ,

    codes.COROSYNC_NOT_RUNNING_ON_NODE: lambda info:
        "{node}: corosync is not running"
        .format(**info)
    ,

    codes.COROSYNC_RUNNING_ON_NODE: lambda info:
        "{node}: corosync is running"
        .format(**info)
    ,

    codes.COROSYNC_QUORUM_GET_STATUS_ERROR: lambda info:
        "Unable to get quorum status: {reason}"
        .format(**info)
    ,

    codes.COROSYNC_QUORUM_SET_EXPECTED_VOTES_ERROR: lambda info:
        "Unable to set expected votes: {reason}"
        .format(**info)
    ,

    codes.COROSYNC_QUORUM_HEURISTICS_ENABLED_WITH_NO_EXEC:
        "No exec_NAME options are specified, so heuristics are effectively "
            "disabled"
    ,

    codes.COROSYNC_CONFIG_RELOADED: "Corosync configuration reloaded",

    codes.COROSYNC_CONFIG_RELOAD_ERROR: lambda info:
        "Unable to reload corosync configuration: {reason}"
        .format(**info)
    ,

    codes.UNABLE_TO_READ_COROSYNC_CONFIG: lambda info:
        "Unable to read {path}: {reason}"
        .format(**info)
    ,

    codes.PARSE_ERROR_COROSYNC_CONF_MISSING_CLOSING_BRACE:
        "Unable to parse corosync config: missing closing brace"
    ,

    codes.PARSE_ERROR_COROSYNC_CONF_UNEXPECTED_CLOSING_BRACE:
        "Unable to parse corosync config: unexpected closing brace"
    ,

    codes.PARSE_ERROR_COROSYNC_CONF:
        "Unable to parse corosync config"
    ,

    codes.COROSYNC_OPTIONS_INCOMPATIBLE_WITH_QDEVICE: lambda info:
        "These options cannot be set when the cluster uses a quorum device: {0}"
        .format(", ".join(sorted(info["options_names"])))
    ,

    codes.QDEVICE_ALREADY_DEFINED:
        "quorum device is already defined"
    ,

    codes.QDEVICE_NOT_DEFINED:
        "no quorum device is defined in this cluster"
    ,

    codes.QDEVICE_REMOVE_OR_CLUSTER_STOP_NEEDED: (
        "You need to stop the cluster or remove qdevice from the cluster to"
        " continue"
    ),

    codes.QDEVICE_CLIENT_RELOAD_STARTED:
        "Reloading qdevice configuration on nodes..."
    ,

    codes.QDEVICE_ALREADY_INITIALIZED: lambda info:
        "Quorum device '{model}' has been already initialized"
        .format(**info)
    ,

    codes.QDEVICE_NOT_INITIALIZED: lambda info:
        "Quorum device '{model}' has not been initialized yet"
        .format(**info)
    ,

    codes.QDEVICE_INITIALIZATION_SUCCESS: lambda info:
        "Quorum device '{model}' initialized"
        .format(**info)
    ,

    codes.QDEVICE_INITIALIZATION_ERROR: lambda info:
        "Unable to initialize quorum device '{model}': {reason}"
        .format(**info)
    ,

    codes.QDEVICE_CERTIFICATE_DISTRIBUTION_STARTED:
        "Setting up qdevice certificates on nodes..."
    ,

    codes.QDEVICE_CERTIFICATE_ACCEPTED_BY_NODE: lambda info:
        "{node}: Succeeded"
        .format(**info)
    ,

    codes.QDEVICE_CERTIFICATE_REMOVAL_STARTED:
        "Removing qdevice certificates from nodes..."
    ,

    codes.QDEVICE_CERTIFICATE_REMOVED_FROM_NODE: lambda info:
        "{node}: Succeeded"
        .format(**info)
    ,

    codes.QDEVICE_CERTIFICATE_IMPORT_ERROR: lambda info:
        "Unable to import quorum device certificate: {reason}"
        .format(**info)
    ,

    codes.QDEVICE_CERTIFICATE_SIGN_ERROR: lambda info:
        "Unable to sign quorum device certificate: {reason}"
        .format(**info)
    ,

    codes.QDEVICE_DESTROY_SUCCESS: lambda info:
        "Quorum device '{model}' configuration files removed"
        .format(**info)
    ,

    codes.QDEVICE_DESTROY_ERROR: lambda info:
        "Unable to destroy quorum device '{model}': {reason}"
        .format(**info)
    ,

    codes.QDEVICE_NOT_RUNNING: lambda info:
        "Quorum device '{model}' is not running"
        .format(**info)
    ,

    codes.QDEVICE_GET_STATUS_ERROR: lambda info:
        "Unable to get status of quorum device '{model}': {reason}"
        .format(**info)
    ,

    codes.QDEVICE_USED_BY_CLUSTERS: lambda info:
        "Quorum device is currently being used by cluster(s): {cluster_list}"
        .format(cluster_list=", ".join(info["clusters"]))
    ,

    codes.CMAN_UNSUPPORTED_COMMAND:
        "This command is not supported on CMAN clusters"
    ,

    codes.ID_ALREADY_EXISTS: lambda info:
        "'{id}' already exists"
        .format(**info)
    ,

    codes.ID_BELONGS_TO_UNEXPECTED_TYPE: id_belongs_to_unexpected_type,

    codes.ID_NOT_FOUND: id_not_found,

    codes.STONITH_RESOURCES_DO_NOT_EXIST: lambda info:
        "Stonith resource(s) '{stonith_id_list}' do not exist"
        .format(
            stonith_id_list="', '".join(info["stonith_ids"]),
            **info
        )
    ,

    codes.CIB_ACL_ROLE_IS_ALREADY_ASSIGNED_TO_TARGET: lambda info:
        "Role '{role_id}' is already asigned to '{target_id}'"
        .format(**info)
    ,

    codes.CIB_ACL_ROLE_IS_NOT_ASSIGNED_TO_TARGET: lambda info:
        "Role '{role_id}' is not assigned to '{target_id}'"
        .format(**info)
    ,

    codes.CIB_ACL_TARGET_ALREADY_EXISTS: lambda info:
        "'{target_id}' already exists"
        .format(**info)
    ,

    codes.CIB_FENCING_LEVEL_ALREADY_EXISTS: lambda info:
        (
            "Fencing level for '{target}' at level '{level}' "
            "with device(s) '{device_list}' already exists"
        ).format(
            device_list=",".join(info["devices"]),
            target=format_fencing_level_target(
                info["target_type"], info["target_value"]
            ),
            **info
        )
    ,

    codes.CIB_FENCING_LEVEL_DOES_NOT_EXIST: lambda info:
        "Fencing level {part_target}{part_level}{part_devices}does not exist"
        .format(
            part_target=(
                "for '{0}' ".format(format_fencing_level_target(
                    info["target_type"], info["target_value"]
                ))
                if info["target_type"] and info["target_value"]
                else ""
            ),
            part_level=format_optional(info["level"], "at level '{0}' "),
            part_devices=format_optional(
                ",".join(info["devices"]) if info["devices"] else "",
                "with device(s) '{0}' "
            )
        )
    ,

    codes.CIB_LOAD_ERROR: "unable to get cib",

    codes.CIB_LOAD_ERROR_SCOPE_MISSING: lambda info:
        "unable to get cib, scope '{scope}' not present in cib"
        .format(**info)
    ,

    codes.CIB_LOAD_ERROR_BAD_FORMAT: lambda info:
       "unable to get cib, {reason}"
       .format(**info)
    ,

    codes.CIB_CANNOT_FIND_MANDATORY_SECTION: lambda info:
        "Unable to get {section} section of cib"
        .format(**info)
    ,

    codes.CIB_PUSH_ERROR: lambda info:
        "Unable to update cib\n{reason}\n{pushed_cib}"
        .format(**info)
    ,

    codes.CIB_DIFF_ERROR: lambda info:
        "Unable to diff CIB: {reason}\n{cib_new}"
        .format(**info)
    ,

    codes.CIB_PUSH_FORCED_FULL_DUE_TO_CRM_FEATURE_SET: lambda info:
        (
            "Replacing the whole CIB instead of applying a diff, a race "
            "condition may happen if the CIB is pushed more than once "
            "simultaneously. To fix this, upgrade pacemaker to get "
            "crm_feature_set at least {required_set}, current is {current_set}."
        ).format(**info)
    ,

    codes.CIB_SAVE_TMP_ERROR: lambda info:
        "Unable to save CIB to a temporary file: {reason}"
        .format(**info)
    ,

    codes.CRM_MON_ERROR:
        "error running crm_mon, is pacemaker running?"
    ,

    codes.BAD_CLUSTER_STATE_FORMAT:
        "cannot load cluster status, xml does not conform to the schema"
    ,

    codes.WAIT_FOR_IDLE_NOT_SUPPORTED:
        "crm_resource does not support --wait, please upgrade pacemaker"
    ,

    codes.WAIT_FOR_IDLE_NOT_LIVE_CLUSTER:
        "Cannot use '-f' together with '--wait'"
    ,

    codes.WAIT_FOR_IDLE_TIMED_OUT: lambda info:
        "waiting timeout\n\n{reason}"
        .format(**info)
    ,

    codes.WAIT_FOR_IDLE_ERROR: lambda info:
        "{reason}"
        .format(**info)
    ,

    codes.RESOURCE_BUNDLE_ALREADY_CONTAINS_A_RESOURCE: lambda info:
        (
            "bundle '{bundle_id}' already contains resource '{resource_id}'"
            ", a bundle may contain at most one resource"
        ).format(**info)
    ,

    codes.RESOURCE_CLEANUP_ERROR: lambda info:
        (
             (
                "Unable to forget failed operations of resource: {resource}"
                "\n{reason}"
             )
             if info["resource"] else
             "Unable to forget failed operations of resources\n{reason}"
        ).format(**info)
    ,

    codes.RESOURCE_REFRESH_ERROR: lambda info:
        (
             "Unable to delete history of resource: {resource}\n{reason}"
             if info["resource"] else
             "Unable to delete history of resources\n{reason}"
        ).format(**info)
    ,

    codes.RESOURCE_REFRESH_TOO_TIME_CONSUMING: lambda info:
        (
             "Deleting history of all resources on all nodes will execute more "
             "than {threshold} operations in the cluster, which may "
             "negatively impact the responsiveness of the cluster. "
             "Consider specifying resource and/or node"
       ).format(**info)
    ,

    codes.RESOURCE_OPERATION_INTERVAL_DUPLICATION: lambda info: (
        "multiple specification of the same operation with the same interval:\n"
        +"\n".join([
            "{0} with intervals {1}".format(name, ", ".join(intervals))
            for name, intervals_list in info["duplications"].items()
            for intervals in intervals_list
        ])
    ),

    codes.RESOURCE_OPERATION_INTERVAL_ADAPTED: lambda info:
        (
            "changing a {operation_name} operation interval"
                " from {original_interval}"
                " to {adapted_interval} to make the operation unique"
        ).format(**info)
    ,

    codes.RESOURCE_RUNNING_ON_NODES:  resource_running_on_nodes,

    codes.RESOURCE_DOES_NOT_RUN: lambda info:
        "resource '{resource_id}' is not running on any node"
        .format(**info)
    ,

    codes.RESOURCE_IS_UNMANAGED: lambda info:
        "'{resource_id}' is unmanaged"
        .format(**info)
    ,

    codes.RESOURCE_IS_GUEST_NODE_ALREADY: lambda info:
        "the resource '{resource_id}' is already a guest node"
        .format(**info)
    ,

    codes.RESOURCE_MANAGED_NO_MONITOR_ENABLED: lambda info:
        (
            "Resource '{resource_id}' has no enabled monitor operations."
            " Re-run with '--monitor' to enable them."
        )
        .format(**info)
    ,

    codes.NODE_NOT_FOUND: lambda info:
        "{desc} '{node}' does not appear to exist in configuration".format(
            desc=build_node_description(info["searched_types"]),
            node=info["node"]
        )
    ,

    codes.NODE_REMOVE_IN_PACEMAKER_FAILED: lambda info:
        "unable to remove node '{node_name}' from pacemaker{reason_part}"
        .format(
            reason_part=format_optional(info["reason"], ": {0}"),
            **info

        )
    ,

    codes.NODE_TO_CLEAR_IS_STILL_IN_CLUSTER: lambda info:
        (
            "node '{node}' seems to be still in the cluster"
            "; this command should be used only with nodes that have been"
            " removed from the cluster"
        )
        .format(**info)
    ,

    codes.MULTIPLE_RESULTS_FOUND: lambda info:
        "multiple {result_type} {search_description} found: {what_found}"
        .format(
            what_found=joined_list(info["result_identifier_list"]),
            search_description="" if not info["search_description"]
                else "for '{0}'".format(info["search_description"])
            ,
            result_type=info["result_type"]
        )
    ,

    codes.PACEMAKER_LOCAL_NODE_NAME_NOT_FOUND: lambda info:
        "unable to get local node name from pacemaker: {reason}"
        .format(**info)
    ,

    codes.RRP_ACTIVE_NOT_SUPPORTED:
        "using a RRP mode of 'active' is not supported or tested"
    ,

    codes.IGNORED_CMAN_UNSUPPORTED_OPTION: lambda info:
        "{option_name} ignored as it is not supported on CMAN clusters"
        .format(**info)
    ,

    codes.NON_UDP_TRANSPORT_ADDR_MISMATCH:
        "--addr0 and --addr1 can only be used with --transport=udp"
    ,

    codes.CMAN_UDPU_RESTART_REQUIRED: (
        "Using udpu transport on a CMAN cluster,"
        " cluster restart is required after node add or remove"
    ),

    codes.CMAN_BROADCAST_ALL_RINGS: (
        "Enabling broadcast for all rings as CMAN does not support"
        " broadcast in only one ring"
    ),

    codes.SERVICE_START_STARTED: partial(service_operation_started, "Starting"),
    codes.SERVICE_START_ERROR: partial(service_operation_error, "start"),
    codes.SERVICE_START_SUCCESS: partial(service_operation_success, "started"),
    codes.SERVICE_START_SKIPPED: partial(service_operation_skipped, "starting"),

    codes.SERVICE_STOP_STARTED: partial(service_operation_started, "Stopping"),
    codes.SERVICE_STOP_ERROR: partial(service_operation_error, "stop"),
    codes.SERVICE_STOP_SUCCESS: partial(service_operation_success, "stopped"),

    codes.SERVICE_ENABLE_STARTED: partial(
        service_operation_started, "Enabling"
    ),
    codes.SERVICE_ENABLE_ERROR: partial(service_operation_error, "enable"),
    codes.SERVICE_ENABLE_SUCCESS: partial(service_operation_success, "enabled"),
    codes.SERVICE_ENABLE_SKIPPED: partial(
        service_operation_skipped, "enabling"
    ),

    codes.SERVICE_DISABLE_STARTED:
        partial(service_operation_started, "Disabling")
     ,
    codes.SERVICE_DISABLE_ERROR: partial(service_operation_error, "disable"),
    codes.SERVICE_DISABLE_SUCCESS: partial(service_operation_success, "disabled"),

    codes.SERVICE_KILL_ERROR: lambda info:
        "Unable to kill {service_list}: {reason}"
        .format(
            service_list=", ".join(info["services"]),
            **info
        )
    ,

    codes.SERVICE_KILL_SUCCESS: lambda info:
        "{services_list} killed"
        .format(
            service_list=", ".join(info["services"]),
            **info
        )
    ,

    codes.UNABLE_TO_GET_AGENT_METADATA: lambda info:
        (
            "Agent '{agent}' is not installed or does not provide valid"
            " metadata: {reason}"
        ).format(**info)
    ,

    codes.INVALID_RESOURCE_AGENT_NAME: lambda info:
        (
            "Invalid resource agent name '{name}'."
            " Use standard:provider:type when standard is 'ocf' or"
            " standard:type otherwise."
            " List of standards and providers can be obtained by using commands"
            " 'pcs resource standards' and 'pcs resource providers'"
        )
        .format(**info)
    ,

    codes.INVALID_STONITH_AGENT_NAME: lambda info:
        (
            "Invalid stonith agent name '{name}'."
            " List of agents can be obtained by using command"
            " 'pcs stonith list'. Do not use the 'stonith:' prefix. Agent name"
            " cannot contain the ':' character."
        )
        .format(**info)
    ,

    codes.AGENT_NAME_GUESS_FOUND_MORE_THAN_ONE: lambda info:
        (
            "Multiple agents match '{agent}'"
            ", please specify full name: {possible_agents_str}"
        ).format(**info)
    ,

    codes.AGENT_NAME_GUESS_FOUND_NONE: lambda info:
        "Unable to find agent '{agent}', try specifying its full name"
        .format(**info)
    ,

    codes.AGENT_NAME_GUESSED: lambda info:
        "Assumed agent name '{guessed_name}' (deduced from '{entered_name}')"
        .format(**info)
    ,

    codes.OMITTING_NODE: lambda info:
        "Omitting node '{node}'"
        .format(**info)
    ,

    codes.SBD_CHECK_STARTED: "Running SBD pre-enabling checks...",

    codes.SBD_CHECK_SUCCESS: lambda info:
        "{node}: SBD pre-enabling checks done"
        .format(**info)
    ,

    codes.SBD_CONFIG_DISTRIBUTION_STARTED: "Distributing SBD config...",

    codes.SBD_CONFIG_ACCEPTED_BY_NODE: lambda info:
        "{node}: SBD config saved"
        .format(**info)
    ,

    codes.UNABLE_TO_GET_SBD_CONFIG: lambda info:
        "Unable to get SBD configuration from node '{node}'{reason_suffix}"
        .format(
            reason_suffix=format_optional(info["reason"], ": {0}"),
            **info
        )
    ,

    codes.SBD_ENABLING_STARTED: lambda info:
        "Enabling SBD service..."
        .format(**info)
    ,

    codes.SBD_DISABLING_STARTED: "Disabling SBD service...",

    codes.SBD_DEVICE_INITIALIZATION_STARTED: lambda info:
        "Initializing device(s) {devices}..."
        .format(devices=", ".join(info["device_list"]))
    ,

    codes.SBD_DEVICE_INITIALIZATION_SUCCESS:
        "Device(s) initialized successfuly",

    codes.SBD_DEVICE_INITIALIZATION_ERROR: lambda info:
        "Initialization of device(s) failed: {reason}"
        .format(**info)
    ,

    codes.SBD_DEVICE_LIST_ERROR: lambda info:
        "Unable to get list of messages from device '{device}': {reason}"
        .format(**info)
    ,

    codes.SBD_DEVICE_MESSAGE_ERROR: lambda info:
        "Unable to set message '{message}' for node '{node}' on device "
        "'{device}'"
        .format(**info)
    ,

    codes.SBD_DEVICE_DUMP_ERROR: lambda info:
        "Unable to get SBD headers from device '{device}': {reason}"
        .format(**info)
    ,

    codes.FILES_DISTRIBUTION_STARTED: lambda info:
        "Sending {description}{where}".format(
            where=(
                "" if not info["node_list"]
                else " to " + joined_list(info["node_list"])
            ),
            description=info["description"] if info["description"]
                else joined_list(info["file_list"])
        )
    ,

    codes.FILE_DISTRIBUTION_SUCCESS: lambda info:
        "{node}: successful distribution of the file '{file_description}'"
        .format(
            **info
        )
    ,


    codes.FILE_DISTRIBUTION_ERROR: lambda info:
        "{node}: unable to distribute file '{file_description}': {reason}"
        .format(
            **info
        )
    ,

    codes.FILES_REMOVE_FROM_NODE_STARTED: lambda info:
        "Requesting remove {description}{where}".format(
            where=(
                "" if not info["node_list"]
                else " from " + joined_list(info["node_list"])
            ),
            description=info["description"] if info["description"]
                else joined_list(info["file_list"])
        )
    ,

    codes.FILE_REMOVE_FROM_NODE_SUCCESS: lambda info:
        "{node}: successful removal of the file '{file_description}'"
        .format(
            **info
        )
    ,


    codes.FILE_REMOVE_FROM_NODE_ERROR: lambda info:
        "{node}: unable to remove file '{file_description}': {reason}"
        .format(
            **info
        )
    ,

    codes.SERVICE_COMMANDS_ON_NODES_STARTED: lambda info:
        "Requesting {description}{where}".format(
            where=(
                "" if not info["node_list"]
                else " on " + joined_list(info["node_list"])
            ),
            description=info["description"] if info["description"]
                else joined_list(info["action_list"])
        )
    ,

    codes.SERVICE_COMMAND_ON_NODE_SUCCESS: lambda info:
        "{node}: successful run of '{service_command_description}'"
        .format(
            **info
        )
    ,

    codes.SERVICE_COMMAND_ON_NODE_ERROR: lambda info:
        (
            "{node}: service command failed:"
            " {service_command_description}: {reason}"
        )
        .format(
            **info
        )
    ,

    codes.SBD_DEVICE_PATH_NOT_ABSOLUTE: lambda info:
        "Device path '{device}'{on_node} is not absolute"
        .format(
            on_node=format_optional(
                info["node"], " on node '{0}'".format(info["node"])
            ),
            **info
        )
    ,

    codes.SBD_DEVICE_DOES_NOT_EXIST: lambda info:
        "{node}: device '{device}' not found"
        .format(**info)
    ,

    codes.SBD_DEVICE_IS_NOT_BLOCK_DEVICE: lambda info:
        "{node}: device '{device}' is not a block device"
        .format(**info)
    ,

    codes.INVALID_RESPONSE_FORMAT: lambda info:
        "{node}: Invalid format of response"
        .format(**info)
    ,

    codes.SBD_NO_DEVICE_FOR_NODE: lambda info:
        "No device defined for node '{node}'"
        .format(**info)
    ,

    codes.SBD_TOO_MANY_DEVICES_FOR_NODE: lambda info:
        (
            "More than {max_devices} devices defined for node '{node}' "
            "(devices: {devices})"
        )
        .format(devices=", ".join(info["device_list"]), **info)
    ,

    codes.SBD_NOT_INSTALLED: lambda info:
        "SBD is not installed on node '{node}'"
        .format(**info)
    ,

    codes.WATCHDOG_NOT_FOUND: lambda info:
        "Watchdog '{watchdog}' does not exist on node '{node}'"
        .format(**info)
    ,

    codes.WATCHDOG_INVALID: lambda info:
        "Watchdog path '{watchdog}' is invalid."
        .format(**info)
    ,

    codes.UNABLE_TO_GET_SBD_STATUS: lambda info:
        "Unable to get status of SBD from node '{node}'{reason_suffix}"
        .format(
            reason_suffix=format_optional(info["reason"], ": {0}"),
            **info
        )
    ,

    codes.CLUSTER_RESTART_REQUIRED_TO_APPLY_CHANGES:
        "Cluster restart is required in order to apply these changes."
    ,

    codes.CIB_ALERT_RECIPIENT_ALREADY_EXISTS: lambda info:
        "Recipient '{recipient}' in alert '{alert}' already exists"
        .format(**info)
    ,

    codes.CIB_ALERT_RECIPIENT_VALUE_INVALID: lambda info:
        "Recipient value '{recipient}' is not valid."
        .format(**info)
    ,

    codes.CIB_UPGRADE_SUCCESSFUL:
        "CIB has been upgraded to the latest schema version."
    ,

    codes.CIB_UPGRADE_FAILED: lambda info:
        "Upgrading of CIB to the latest schema failed: {reason}"
        .format(**info)
    ,

    codes.CIB_UPGRADE_FAILED_TO_MINIMAL_REQUIRED_VERSION: lambda info:
        (
            "Unable to upgrade CIB to required schema version"
            " {required_version} or higher. Current version is"
            " {current_version}. Newer version of pacemaker is needed."
        )
        .format(**info)
    ,

    codes.FILE_ALREADY_EXISTS: lambda info:
        "{_node}{_file_role} file '{file_path}' already exists"
        .format(
            _node=format_optional(info["node"], NODE_PREFIX),
            _file_role=format_file_role(info["file_role"]),
            **info
        )
    ,

    codes.FILE_DOES_NOT_EXIST: lambda info:
        "{_file_role} file '{file_path}' does not exist"
        .format(
            _file_role=format_file_role(info["file_role"]),
            **info
        )
    ,

    codes.FILE_IO_ERROR: lambda info:
        "Unable to {operation} {_file_role}{_file_path}: {reason}"
        .format(
            _file_path=format_optional(info["file_path"], " '{0}'"),
            _file_role=format_file_role(info["file_role"]),
            **info
        )
    ,

    codes.UNABLE_TO_DETERMINE_USER_UID: lambda info:
        "Unable to determine uid of user '{user}'"
        .format(**info)
    ,

    codes.UNABLE_TO_DETERMINE_GROUP_GID: lambda info:
        "Unable to determine gid of group '{group}'"
        .format(**info)
    ,

    codes.UNSUPPORTED_OPERATION_ON_NON_SYSTEMD_SYSTEMS:
        "unsupported operation on non systemd systems"
    ,

    codes.LIVE_ENVIRONMENT_REQUIRED: lambda info:
        "This command does not support {forbidden_options}"
        .format(
            forbidden_options=joined_list(info["forbidden_options"], {
                "BOOTH_CONF": "--booth-conf",
                "BOOTH_KEY": "--booth-key",
                "CIB": "-f",
                "COROSYNC_CONF": "--corosync_conf",
            })
        )
    ,

    codes.LIVE_ENVIRONMENT_REQUIRED_FOR_LOCAL_NODE:
        "Node(s) must be specified if -f is used"
    ,

    codes.NOLIVE_SKIP_FILES_DISTRIBUTION: lambda info:
        (
            "the distribution of {files} to {nodes} was skipped because command"
            " does not run on live cluster (e.g. -f was used)."
            " You will have to do it manually."
        ).format(
            files=joined_list(info["files_description"]),
            nodes=joined_list(info["nodes"]),
        )
    ,
    codes.NOLIVE_SKIP_FILES_REMOVE: lambda info:
        (
            "{files} remove from {nodes} was skipped because command"
            " does not run on live cluster (e.g. -f was used)."
            " You will have to do it manually."
        ).format(
            files=joined_list(info["files_description"]),
            nodes=joined_list(info["nodes"]),
        )
    ,
    codes.NOLIVE_SKIP_SERVICE_COMMAND_ON_NODES: lambda info:
        (
            "running '{command}' on {nodes} was skipped"
                " because command does not run on live cluster (e.g. -f was"
                " used). You will have to run it manually."
        ).format(
            command="{0} {1}".format(info["service"], info["command"]),
            nodes=joined_list(info["nodes"]),
        )
    ,

    codes.COROSYNC_QUORUM_CANNOT_DISABLE_ATB_DUE_TO_SBD: lambda info:
        "unable to disable auto_tie_breaker: SBD fencing will have no effect"
        .format(**info)
    ,

    codes.SBD_REQUIRES_ATB: lambda info:
        "auto_tie_breaker quorum option will be enabled to make SBD fencing "
        "effective. Cluster has to be offline to be able to make this change."
    ,

    codes.CLUSTER_CONF_LOAD_ERROR_INVALID_FORMAT: lambda info:
        "unable to get cluster.conf: {reason}"
        .format(**info)
    ,

    codes.CLUSTER_CONF_READ_ERROR: lambda info:
        "Unable to read {path}: {reason}"
        .format(**info)
    ,
    codes.USE_COMMAND_NODE_ADD_REMOTE: lambda info:
        (
            "this command is not sufficient for creating a remote connection,"
            " use 'pcs cluster node add-remote'"
        )
    ,
    codes.USE_COMMAND_NODE_ADD_GUEST: lambda info:
        (
            "this command is not sufficient for creating a guest node, use"
            " 'pcs cluster node add-guest'"
        )
    ,
    codes.USE_COMMAND_NODE_REMOVE_GUEST: lambda info:
        (
            "this command is not sufficient for removing a guest node, use"
            " 'pcs cluster node remove-guest'"
        )
    ,

    codes.TMP_FILE_WRITE: lambda info:
        (
            "Writing to a temporary file {file_path}:\n"
            "--Debug Content Start--\n{content}\n--Debug Content End--\n"
        ).format(**info)
    ,
    codes.UNABLE_TO_PERFORM_OPERATION_ON_ANY_NODE:
        "Unable to perform operation on any available node/host, therefore it "
        "is not possible to continue."
    ,
    codes.SBD_LIST_WATCHDOG_ERROR: lambda info:
        "Unable to query available watchdogs from sbd: {reason}".format(**info)
    ,
    codes.SBD_WATCHDOG_NOT_SUPPORTED: lambda info:
        (
            "{node}: Watchdog '{watchdog}' is not supported (it may be a "
            "software watchdog)"
        ).format(**info)
    ,
    codes.SBD_WATCHDOG_VALIDATION_INACTIVE:
        "Not validating the watchdog"
    ,
    codes.SBD_WATCHDOG_TEST_ERROR: lambda info:
        "Unable to initialize test of the watchdog: {reason}".format(**info)
    ,
    codes.SBD_WATCHDOG_TEST_MULTUPLE_DEVICES:
        "Multiple watchdog devices available, therefore, watchdog which should "
        "be tested has to be specified. To list available watchdog devices use "
        "command 'pcs stonith sbd watchdog list'"
    ,
    codes.SBD_WATCHDOG_TEST_FAILED:
        "System should have been reset already"
    ,
    codes.SYSTEM_WILL_RESET:
        "System will reset shortly"
    ,
    codes.RESOURCE_IN_BUNDLE_NOT_ACCESSIBLE: lambda info:
        (
            "Resource '{resource_id}' will not be accessible by the cluster "
            "inside bundle '{bundle_id}'. At least one of bundle options "
            "'control-port' or 'ip-range-start' has to be specified."
        ).format(**info)
}



from __future__ import (
    absolute_import,
    division,
    print_function,
)

from pcs.lib.cib.resource import (
    bundle,
    clone,
    common,
    group,
    guest_node,
    operations,
    primitive,
    remote_node,
)
